"""
Sample Docopt script for bosh import testing

Usage:
  test_import X a <list>...
  test_import X a b g
  test_import X b (--bar1=Number|--bar2=File) [--foo1=String|--foo2=<Number>|--foo3=File]
  test_import X b g (--bar1=Number|--bar2=File)
  test_import X c --d1=String --d2=Number --d3=File --d4=Flag --d5=<dummy>
  test_import Y <m1> <m2> <m31> <m41> [--desc1]
  test_import Y <m1> <m2> <m32> <m42> --desc2
  test_import -h | --help

Options:
  -h --help       Help descriptions
  --bar1=Number   requires b
  --bar2=File     requires b
  --foo1=String   requires bar1_bar2
  --foo2=<Number>   requires bar1_bar2
  --foo3=File     requires bar1_bar2
  --d1=String     string type
  --d2=Number     number type [default: 50]
  --d3=File       file type
  --d4=Flag       flag type
  --d5=<dummy>    non-supported type
  --desc1         desc1 d3$c&p+0s ~!@#$%^&*()_+=-
  --desc2         desc2 descriptions
"""



import os
import urllib

import requests

from boutiques.logger import print_info, raise_error
from boutiques.searcher import Searcher
from boutiques.zenodoHelper import ZenodoError, ZenodoHelper

try:
    # Python 3
    from urllib.request import urlopen, urlretrieve
except ImportError:
    # Python 2
    from urllib import urlretrieve

    from urllib2 import urlopen


class Puller:

    def __init__(self, zids, verbose=False, sandbox=False):
        # remove zenodo prefix
        self.zenodo_entries = []
        self.cache_dir = os.path.join(
            os.path.expanduser("~"),
            ".cache",
            "boutiques",
            "sandbox" if sandbox else "production",
        )
        discarded_zids = zids
        # This removes duplicates, should maintain order
        zids = list(dict.fromkeys(zids))
        for zid in zids:
            discarded_zids.remove(zid)
            try:
                # Zenodo returns the full DOI, but for the purposes of
                # Boutiques we just use the Zenodo-specific portion (as its the
                # unique part). If the API updates on Zenodo to no longer
                # provide the full DOI, this still works because it just grabs
                # the last thing after the split.
                zid = zid.split("/")[-1]
                newzid = zid.split(".", 1)[1]
                newfname = os.path.join(self.cache_dir, f"zenodo-{newzid}.json")
                self.zenodo_entries.append({"zid": newzid, "fname": newfname})
            except IndexError:
                raise_error(
                    ZenodoError,
                    "Zenodo ID must be prefixed by " "'zenodo', e.g. zenodo.123456",
                )
        self.verbose = verbose
        self.sandbox = sandbox
        if self.verbose:
            for zid in discarded_zids:
                print_info(f"Discarded duplicate id {zid}")
        self.zenodo_helper = ZenodoHelper(sandbox=self.sandbox, verbose=self.verbose)

    def pull(self):
        # return cached file if it exists
        json_files = []
        for entry in self.zenodo_entries:
            if os.path.isfile(entry["fname"]):
                if self.verbose:
                    print_info(f"Found cached file at {entry['fname']}")
                json_files.append(entry["fname"])
                continue

            searcher = Searcher(
                entry["zid"], self.verbose, self.sandbox, exact_match=True
            )
            r = self.zenodo_helper.zenodo_search(searcher.query, searcher.query_line)
            if not len(r.json()["hits"]["hits"]):
                raise_error(
                    ZenodoError,
                    f"Descriptor \"{entry['zid']}\" not found",
                )
            for hit in r.json()["hits"]["hits"]:
                file_path = hit["files"][0]["links"]["self"]
                file_name = file_path.split(os.sep)[-1]
                if hit["id"] == int(entry["zid"]):
                    if not os.path.exists(self.cache_dir):
                        os.makedirs(self.cache_dir)
                    if self.verbose:
                        print_info(f"Downloading descriptor {file_name}")
                    downloaded = urlretrieve(file_path, entry["fname"])
                    if self.verbose:
                        print_info("Downloaded descriptor to " + downloaded[0])
                    json_files.append(downloaded[0])
                else:
                    raise_error(
                        ZenodoError,
                        'Searched-for descriptor "{}" '
                        'does not match descriptor "{}" returned '
                        "from Zenodo".format(entry["zid"], hit["id"]),
                    )

        return json_files



#!/usr/bin/env python

import os
import subprocess
import sys
from unittest import mock

import pytest
from boutiques_mocks import *

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.localExec import ExecutorError
from boutiques.tests.BaseTest import BaseTest


def mock_mkdir():
    return [OSError("cannot mkdir"), None]


def mock_mkdir_timeout():
    return OSError("cannot mkdir")


def mock_sleep():
    return None


def mock_clean_up():
    return None


def mock_exists():
    return [False, True]


def mock_sing_pull():
    return ((None, None), 0), None


def mock_os_rename():
    return None


class TestPrepare(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("example1")

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_prepare_docker(self):
        ret = bosh.execute("prepare", self.example1_descriptor)
        self.assertIn("Pulled from Docker", ret.stdout)

    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_prepare_sing(self):
        ret = bosh.execute("prepare", self.get_file_path("example1_sing.json"))
        self.assertIn("Local (boutiques-example1-test.simg)", ret.stdout)

    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_prepare_sing_specify_imagepath(self):
        ret = bosh.execute(
            "prepare",
            self.get_file_path("example1_sing.json"),
            "--imagepath",
            os.path.join(os.getcwd(), "boutiques-example1-test.simg"),
        )
        self.assertIn("Local (boutiques-example1-test.simg)", ret.stdout)
        self.assertNotIn("SINGULARITY_PULLFOLDER", os.environ)

    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_prepare_sing_specify_imagepath_basename_only(self):
        ret = bosh.execute(
            "prepare",
            self.get_file_path("example1_sing.json"),
            "--imagepath",
            "boutiques-example1-test.simg",
        )
        self.assertIn("Local (boutiques-example1-test.simg)", ret.stdout)
        self.assertNotIn("SINGULARITY_PULLFOLDER", os.environ)

    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_prepare_sing_image_does_not_exist(self):
        # Specify the wrong path for the image
        # Will try to pull it and pull will fail
        with pytest.raises(ExecutorError) as e:
            bosh.execute(
                "prepare",
                self.get_file_path("example1_sing_crash_pull.json"),
                "--imagepath",
                os.path.join(os.path.expanduser("~"), "boutiques-example1-latest.simg"),
            )
        self.assertIn("Could not pull Singularity image", str(e.getrepr(style="long")))

    @mock.patch("os.mkdir", side_effect=mock_mkdir())
    @mock.patch(
        "boutiques.localExec.LocalExecutor._singConExists",
        side_effect=mock_exists(),
    )
    @mock.patch(
        "boutiques.localExec.LocalExecutor._cleanUpAfterSingPull",
        side_effect=mock_clean_up(),
    )
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_prepare_sing_multiple_processes(
        self, mock_mkdir, mock_exists, mock_clean_up
    ):
        # Specify path for image that does not exist.
        # Mock that another process created the image at that path
        # while this process was waiting.
        ret = bosh.execute(
            "prepare",
            self.get_file_path("example1_sing.json"),
            "--imagepath",
            os.path.join(os.path.expanduser("~"), "boutiques-example1-test.simg"),
        )
        self.assertIn("Local (boutiques-example1-test.simg)", ret.stdout)

    @mock.patch("os.mkdir", side_effect=mock_mkdir_timeout())
    @mock.patch("time.sleep", side_effect=mock_sleep())
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_prepare_sing_timeout(self, mock_mkdir, mock_sleep):
        # Specify path for image that does not exist.
        # Mock that another process has the lockdir and this one
        # times out while waiting.
        with pytest.raises(ExecutorError) as e:
            bosh.execute(
                "prepare",
                self.get_file_path("example1_sing.json"),
                "--imagepath",
                os.path.join(os.path.expanduser("~"), "boutiques-example1-test.simg"),
            )
        self.assertIn(
            "Unable to retrieve Singularity image",
            str(e.getrepr(style="long")),
        )

    @mock.patch("os.mkdir", side_effect=mock_mkdir_timeout())
    @mock.patch("time.sleep", side_effect=mock_sleep())
    @mock.patch(
        "boutiques.localExec.LocalExecutor._singConExists",
        side_effect=mock_exists(),
    )
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_prepare_sing_timeout_success(self, mock_mkdir, mock_sleep, mock_exists):
        # Specify path for image that does not exist.
        # Mock that another process has the lockdir and this one
        # times out while waiting, but image was created by the
        # other process.
        ret = bosh.execute(
            "prepare",
            self.get_file_path("example1_sing.json"),
            "--imagepath",
            os.path.join(os.path.expanduser("~"), "boutiques-example1-test.simg"),
        )
        self.assertIn("Local (boutiques-example1-test.simg)", ret.stdout)

    @mock.patch(
        "boutiques.localExec.LocalExecutor._localExecute",
        side_effect=mock_sing_pull(),
    )
    @mock.patch("os.rename", side_effect=mock_os_rename())
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_prepare_sing_pull_success(self, mock_sing_pull, mock_os_rename):
        # Specify path for image that does not exist.
        # Try to pull it and mock that the pull was successful.
        ret = bosh.execute(
            "prepare",
            self.get_file_path("example1_sing_crash_pull.json"),
            "--imagepath",
            os.path.join(os.path.expanduser("~"), "boutiques-example1-test.simg"),
        )
        self.assertIn(
            "Pulled from " "docker://boutiques/example1crashcrashcrash:test",
            ret.stdout,
        )

    def test_prepare_no_container(self):
        self.setup("exec")
        ret = bosh.execute("prepare", self.get_file_path("no_container.json"))
        self.assertIn("Descriptor does not specify a container image.", ret.stdout)



#!/usr/bin/env python

import textwrap
from argparse import ArgumentParser, RawTextHelpFormatter

import simplejson as json
from tabulate import tabulate


class PrettyPrinter:
    # Prints descriptor help text in a pretty format by first creating an
    # argument parser, and then using their help text formatting.
    def __init__(self, descriptor):
        self.sep = "".join(["="] * 80)
        self.desc = descriptor
        self.epilog = ""
        self.createHelpText()

    def createHelpText(self):
        self.createLUT()
        self.descMetadata()

        # Add container information - if applicable
        if self.desc.get("container-image"):
            self.descContainer()

        # Add system requirements - if applicable
        if self.desc.get("suggested-resources"):
            self.descResources()

        # Add error codes - if applicable
        if self.desc.get("error-codes"):
            self.descErrors()

        # Add group information - if applicable
        if self.desc.get("groups"):
            self.descGroups()

        # Add output information - if applicable
        if self.desc.get("output-files"):
            self.descOutputs()

        self.descInputs()
        self.parser.epilog = self.epilog
        self.docstring = self.parser.format_help()
        self.docstring = "\n\n".join(self.docstring.split("\n\n")[1:])

    def createLUT(self):
        # Creates two look-up tables: from ids to value-keys, and the reverse
        if self.desc.get("output-files"):
            self.CLfields = self.desc["inputs"] + self.desc["output-files"]
        else:
            self.CLfields = self.desc["inputs"]

        tmplut = {inp["id"]: inp.get("value-key") for inp in self.CLfields}
        self.lut = {
            inp.get("value-key"): [
                t for t in tmplut.keys() if tmplut[t] == inp.get("value-key")
            ]
            for inp in self.CLfields
            if inp.get("value-key")
        }

    def descMetadata(self):
        # Gather main description and basic metadata
        name = f"Tool name: {self.desc['name']} (ver: {self.desc.get('tool-version')})"
        description = f"Tool description: {self.desc['description']}"
        if self.desc.get("tags"):
            taglist = [
                (f"{k}: {v}" if not isinstance(v, list) else f"{k}: {', '.join(v)}")
                for k, v in self.desc["tags"].items()
            ]
            tags = f"Tags: {'; '.join(taglist)}"
        else:
            tags = ""

        # Grabs command-line, and wraps it right to the first element of it
        # (usually the executable) unless it's too long
        cline = self.desc["command-line"]
        cend = min(len(cline.split(" ")[0]) + 1, 35)
        cline = textwrap.wrap(
            "  " + self.desc["command-line"],
            subsequent_indent="  " + " " * cend,
        )
        cline = "\\n".join(cline)
        cline = f"Command-line:\n{cline}"

        # Initialize the tool description with the pieces we've collected so far
        self.helptext = "{0}\n\n{1}\n{2}\n{3}\n\n{4}\n\n{0}" "".format(
            self.sep, name, description, tags, cline
        )

    def descContainer(self):
        conimage = self.desc["container-image"]
        container_info = "Container Information:\n"
        container_info += "\t"
        container_info += "\n\t".join(
            [f"{k.title()}: {conimage[k]}" for k in conimage.keys()]
        )
        self._addSegment(container_info)

    def descOutputs(self):
        outputs = self.desc["output-files"]
        output_info = "Output Files:"
        config_info = "Config Files:"
        for output in outputs:
            required = "Optional" if output.get("optional") else "Required"
            temp_info = f"\n  Name: {output['name']} ({required})"
            temp_info += f"\n\tFormat: {output['path-template']}"

            # Identifies input dependencies based on filename
            depids = [
                "/".join(self.lut[inp])
                for inp in self.lut.keys()
                if inp in output["path-template"]
            ]
            if depids:
                temp_info += f"\n\tFilename depends on Input IDs: {', '.join(depids)}"

            # Gets stripped extensions
            if output.get("path-template-stripped-extensions"):
                exts = ", ".join(output["path-template-stripped-extensions"])
                temp_info += f"\n\tStripped extensions (before substitution): {exts}"

            # If a config file, add the template
            if output.get("file-template"):
                temp_info += "\n\tTemplate:\n\t {}" "".format(
                    "\n\t ".join(output["file-template"])
                )
                config_info += temp_info
            else:
                output_info += temp_info

        self.epilog = (
            f"\n\n{self.sep}\n\n{config_info}" if config_info != "Config Files:" else ""
        )
        self.epilog += f"\n\n{self.sep}\n\n{output_info}"

    def descGroups(self):
        groups = self.desc["groups"]
        gtypes = ["Mutually Exclusive", "All or None", "One is Required"]
        group_info = "Input Groups:"
        for group in groups:
            gtype = []
            gtype += [0] if bool(group.get("mutually-exclusive")) else []
            gtype += [1] if bool(group.get("all-or-none")) else []
            gtype += [2] if bool(group.get("one-is-required")) else []
            group_info += f"\n\tName: {group['name'].title()}"
            group_info += f"\n\tType: {', '.join([gtypes[ind] for ind in gtype])}\n"
            group_info += f"\tGroup Member IDs: {', '.join(group['members'])}"
            group_info += "\n"
        self._addSegment(group_info)

    def descResources(self):
        res = self.desc["suggested-resources"]
        res_info = "Suggested Resources:"
        for rkey in res.keys():
            kx = rkey.replace("-", " ").title()
            res_info += f"\n\t{kx}: {res[rkey]}"
            if rkey == "ram":
                res_info += " GB"
            elif rkey == "walltime-estimate":
                res_info += " s"
        res_info += "\n"
        self._addSegment(res_info)

    def descErrors(self):
        ecod = self.desc["error-codes"]
        ecod_info = "Error Codes:"
        for ecod_obj in ecod:
            ecod_info += "\n\tReturn Code: {}\n\tDescription: {}\n" "".format(
                ecod_obj["code"], ecod_obj["description"]
            )
        ecod_info += "\n"
        self._addSegment(ecod_info)

    def descInputs(self):
        self.parser = ArgumentParser(
            description=self.helptext,
            formatter_class=RawTextHelpFormatter,
            add_help=False,
        )
        inputs = self.CLfields
        required = self.parser.add_argument_group("required arguments")

        # For every command-line key (i.e. input)...
        for clkey in self.lut.keys():
            # Re-initialize an empty argument
            inp_args = []
            inp_kwargs = {}
            opt_inp_descr = ""
            req_inp_descr = ""
            opt_inp_desc_header = ""
            opt_inp_desc_footer = ""
            req_inp_desc_header = ""
            req_inp_desc_footer = ""

            # Get all inputs with the command-line key
            inps = [inps for inps in inputs if inps.get("value-key") == clkey]

            if not len(inps):
                continue

            # Determine if input is optional or required
            tinp = inps[0]
            cflag = tinp.get("command-line-flag")
            if cflag:
                # argparse crashes when dest is supplied and
                # argument doesn't start with '--'
                if not cflag.startswith("-"):
                    cflag = "--" + cflag
                inp_args += [cflag]
                inp_kwargs["dest"] = clkey
            else:
                inp_args += [clkey]

            # If multiple inputs of the same optionality
            # share a command-line key
            multi_opt = sum(bool(inp.get("optional")) for inp in inps) > 1
            multi_req = sum(not bool(inp.get("optional")) for inp in inps) > 1

            if multi_opt:
                opt_inp_descr += "Multiple Options...\n"
                opt_inp_desc_header = "Option {0}:\n"
                opt_inp_desc_footer = "\n"
            if multi_req:
                req_inp_descr += "Multiple Options...\n"
                req_inp_desc_header = "Option {0}:\n"
                req_inp_desc_footer = "\n"

            # For every input with the clkey (usually just 1)...
            for i_inp, inp in enumerate(inps):
                if bool(inp.get("optional")):
                    opt_inp_descr += opt_inp_desc_header.format(i_inp + 1)
                else:
                    req_inp_descr += req_inp_desc_header.format(i_inp + 1)

                # Grab basic input fields first
                tmp_inp_descr = (
                    "ID: {}\nValue Key: {}\nType: {}\n"
                    "List: {}\nOptional: {}\n"
                    "".format(
                        inp.get("id"),
                        inp.get("value-key"),
                        inp.get("type"),
                        bool(inp.get("list")),
                        bool(inp.get("optional")),
                    )
                )

                # If it's a list, get min and max length and present it sensibly
                if inp.get("list"):
                    milen = inp.get("min-list-entries")
                    malen = inp.get("max-list-entries")
                    if milen and malen:
                        listlen = f"{milen} - {malen}"
                    elif milen:
                        listlen = f">= {milen}"
                    elif malen:
                        listlen = f"<= {malen}"
                    else:
                        listlen = "N/A"
                    tmp_inp_descr += f"List Length: {listlen}\n"

                # If it's a number, get min and max values and exclusivity and
                # present it as a standard number range. Also identify if an Int
                if inp.get("type") == "Number":
                    tmp_inp_descr += f"Integer: {bool(inp.get('integer'))}\n"
                    minum = inp.get("minimum") or "N/A"
                    manum = inp.get("maximum") or "N/A"
                    emi = "(" if inp.get("exclusive-minimum") else "["
                    ema = ")" if inp.get("exclusive-maximum") else "]"
                    tmp_inp_descr += f"Range: {emi}{minum}, {manum}{ema}\n"

                # If there are options, add this to the parser and the default
                # to the description
                if inp.get("value-choices"):
                    inp_kwargs["choices"] = inp["value-choices"]
                if inp.get("default-value"):
                    inp_kwargs["default"] = inp["default-value"]
                    tmp_inp_descr += f"Default Value: {inp['default-value']}\n"

                # Show exclusivity with other inputs
                if inp.get("disables-inputs"):
                    tmp_inp_descr += f"Disables: {', '.join(inp['disables-inputs'])}\n"
                if inp.get("requires-inputs"):
                    tmp_inp_descr += f"Requires: {', '.join(inp['requires-inputs'])}\n"

                # Show exclusivity of values with other inputs
                if inp.get("value-disables") and inp.get("value-requires"):
                    tmp_table_headers = ["Value", "Disables", "Requires"]
                    tmp_table = []
                    vdtab = inp["value-disables"]
                    vrtab = inp["value-requires"]
                    for tkey in vdtab:
                        tmp_table += [
                            [
                                tkey,
                                ", ".join(vdtab[tkey]),
                                ", ".join(vrtab[tkey]),
                            ]
                        ]
                    tmp_table = tabulate(tmp_table, headers=tmp_table_headers)
                    tmp_inp_descr += f"Value Dependency: \n {tmp_table}\n"

                # Finally, add the actual description
                if inp.get("description"):
                    descr_text = f"Description: {inp['description']}"
                else:
                    descr_text = ""

                if bool(inp.get("optional")):
                    opt_inp_descr += (
                        tmp_inp_descr
                        + textwrap.fill(descr_text, subsequent_indent=" ")
                        + opt_inp_desc_footer
                    )
                else:
                    req_inp_descr += (
                        tmp_inp_descr
                        + textwrap.fill(descr_text, subsequent_indent=" ")
                        + req_inp_desc_footer
                    )

            # Add args for required inputs
            if req_inp_descr != "":
                inp_kwargs["help"] = textwrap.dedent(req_inp_descr)
                required.add_argument(*inp_args, **inp_kwargs)

            # Add args for optional inputs
            if opt_inp_descr != "":
                inp_kwargs["help"] = textwrap.dedent(opt_inp_descr)
                parsed_flags = list(self.parser._option_string_actions.keys())
                for i in range(0, len(parsed_flags)):
                    if "_DUP" in parsed_flags[i]:
                        parsed_flags[i] = parsed_flags[i][0:-5]
                if inp_args[0] in parsed_flags:
                    inp_args[0] = f"{inp_args[0]}_DUP{parsed_flags.count(inp_args[0])}"
                self.parser.add_argument(*inp_args, **inp_kwargs)

    def _addSegment(self, segment):
        self.helptext += f"\n\n{segment}\n\n{self.sep}"



#!/usr/bin/env python
import os
from codecs import open

import simplejson as json

from boutiques.logger import print_info, raise_error
from boutiques.util.utils import importCatcher


class NexusError(Exception):
    pass


class NexusHelper:

    # Constructor
    def __init__(self, sandbox=False, no_int=False, verbose=False):
        try:
            import nexussdk as nexus
        except ImportError:
            raise_error(
                NexusError,
                "Cannot import nexussdk, " "consider upgrading to python 3.5 or higher",
            )
            return

        self.nexus = nexus
        self.sandbox = sandbox
        self.no_int = no_int
        self.verbose = verbose
        self.config_file = self.get_config_file()
        self.nexus_endpoint = self.get_nexus_endpoint()

    def verify_nexus_input(self, input_token, input_org, input_project):
        access_token = input_token
        org = input_org
        project = input_project
        if access_token is None:
            access_token = self.get_nexus_access_token()
        if org is None:
            org = self.get_nexus_organization()
        if project is None:
            project = self.get_nexus_project()
        self.nexus_test_api(access_token, org, project)
        self.save_nexus_inputs(access_token, org, project)
        return access_token, org, project

    def get_nexus_access_token(self):
        json_creds = self.read_credentials()
        if json_creds.get(self.config_token_property_name()):
            return json_creds.get(self.config_token_property_name())
        if self.no_int:
            raise_error(NexusError, "Cannot find Nexus credentials.")
        prompt = (
            "Please enter your Nexus access token (it will be "
            "saved in {} for future use): ".format(self.config_file)
        )
        return self.prompt(prompt)

    def get_nexus_organization(self):
        json_creds = self.read_credentials()
        if json_creds.get("nexus-organization"):
            return json_creds.get("nexus-organization")
        if self.no_int:
            raise_error(NexusError, "Cannot find Nexus organization.")
        prompt = (
            "Please enter the Nexus organization you want to publish to"
            " (it will be saved in {} for future use): ".format(self.config_file)
        )
        return self.prompt(prompt)

    def get_nexus_project(self):
        json_creds = self.read_credentials()
        if json_creds.get("nexus-project"):
            return json_creds.get("nexus-project")
        if self.no_int:
            raise_error(NexusError, "Cannot find Nexus project.")
        prompt = (
            "Please enter the Nexus project you want to publish to"
            " (it will be saved in {} for future use): ".format(self.config_file)
        )
        return self.prompt(prompt)

    def save_nexus_inputs(self, access_token, org, project):
        json_creds = self.read_credentials()
        json_creds[self.config_token_property_name()] = access_token
        json_creds["nexus-organization"] = org
        json_creds["nexus-project"] = project
        with open(self.config_file, "w") as f:
            f.write(json.dumps(json_creds, indent=4, sort_keys=True))
        if self.verbose:
            print_info(
                "Nexus access token, organization and project"
                " saved in {}".format(self.config_file)
            )

    def get_nexus_endpoint(self):
        # change once nexus is setup
        endpoint = (
            "https://sandbox.bluebrainnexus.io/v1"
            if self.sandbox
            else "https://sandbox.bluebrainnexus.io/v1"
        )
        if self.verbose:
            print_info(f"Using Nexus endpoint {endpoint}")
        return endpoint

    def config_token_property_name(self):
        if self.sandbox:
            return "nexus-access-token-test"
        return "nexus-access-token"

    def read_credentials(self):
        try:
            with open(self.config_file) as f:
                json_creds = json.load(f)
        except OSError:
            json_creds = {}
        except ValueError:
            json_creds = {}
        return json_creds

    @importCatcher()
    def nexus_test_api(self, access_token, org, project):
        from requests.exceptions import ConnectionError, HTTPError

        # Test connection to endpoint without token
        self.nexus.config.set_environment(self.nexus_endpoint)
        try:
            self.nexus.permissions.fetch()
        except ConnectionError as e:
            raise_error(NexusError, "Cannot access Nexus endpoint", e.response)
        if self.verbose:
            print_info("Nexus endpoint is accessible")

        # Test authentication with token, org and project
        self.nexus.config.set_token(access_token)
        try:
            self.nexus.projects.fetch(org, project)
            if self.verbose:
                print_info("Authentication to Nexus successful")
        except HTTPError as e:
            if 404 == e.response.status_code:
                raise_error(
                    NexusError,
                    "No project '{}' in organization '{}' "
                    "in Nexus repository".format(project, org),
                    e.response,
                )
            elif 401 == e.response.status_code:
                raise_error(
                    NexusError,
                    "Cannot authenticate to Nexus API, check " "your access token",
                    e.response,
                )
            else:
                raise_error(NexusError, "Something went wrong", e.response)

    def get_config_file(self):
        return os.path.join(os.path.expanduser("~"), ".nexus")

    def prompt(self, prompt):
        try:
            return raw_input(prompt)  # Python 2
        except NameError:
            return input(prompt)  # Python 3

    def publish(self, org_label, project_label, filepath):
        self.nexus.files.create(org_label, project_label, filepath)



#!/usr/bin/env python

import os
import os.path as op
import sys

import pytest
from six import string_types

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.tests.BaseTest import BaseTest


class TestPPrint(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("pprint")

    def test_doesntcrash(self):
        fil = self.get_file_path("test_pretty_print.json")
        prettystring = bosh.pprint(fil)
        self.assertIsInstance(prettystring, (str,))

    def test_categories_and_order(self):
        fil = self.get_file_path("test_pretty_print.json")
        prettystring = bosh.pprint(fil)
        i_tl_descs = prettystring.index("Tool name")
        i_con_info = prettystring.index("Container Information")
        i_sugg_res = prettystring.index("Suggested Resources")
        i_er_codes = prettystring.index("Error Codes")
        i_inp_grps = prettystring.index("Input Groups")
        i_pos_args = prettystring.index("positional arguments")
        if sys.version_info.minor < 10:
            i_opt_args = prettystring.index("optional arguments")
        else:
            i_opt_args = prettystring.index("options")
        i_req_args = prettystring.index("required arguments")
        i_conf_fil = prettystring.index("Config Files")
        i_out_file = prettystring.index("Output Files")

        # Fluff asserts for easy debugging
        self.assertTrue(i_tl_descs < i_con_info)
        self.assertTrue(i_con_info < i_sugg_res)
        self.assertTrue(i_sugg_res < i_er_codes)
        self.assertTrue(i_er_codes < i_inp_grps)
        self.assertTrue(i_inp_grps < i_pos_args)
        self.assertTrue(i_pos_args < i_opt_args)
        self.assertTrue(i_opt_args < i_req_args)
        self.assertTrue(i_req_args < i_conf_fil)
        self.assertTrue(i_conf_fil < i_out_file)

    def test_input_optionality_separation(self):
        fil = self.get_file_path("test_pretty_print.json")
        prettystring = bosh.pprint(fil)
        if sys.version_info.minor < 10:
            inputs = prettystring.split("=" * 80)[6].split("arguments:")
            positional_inputs = inputs[1]
            optional_inputs = inputs[2]
            required_inputs = inputs[3]
            self.assertFalse("Optional: False" in positional_inputs)
            self.assertFalse("Optional: False" in optional_inputs)
            self.assertFalse("Optional: True" in required_inputs)
        else:
            inputs = prettystring.split("=" * 80)[6].split("arguments:")
            positional_inputs = inputs[1]
            required_inputs = inputs[2]
            self.assertFalse("Optional: False" in positional_inputs)
            self.assertFalse("Optional: True" in required_inputs)
            inputs = prettystring.split("=" * 80)[6].split("options:")
            optional_inputs = inputs[0]
            self.assertFalse("Optional: False" in optional_inputs)

    def test_output_config_separation(self):
        fil = self.get_file_path("test_pretty_print.json")
        prettystring = bosh.pprint(fil)
        categories = prettystring.split("=" * 80)
        configs = categories[7]
        outputs = categories[8]
        self.assertTrue("Config Files:" in configs)
        self.assertTrue("Template:" in configs)
        self.assertTrue("Output Files:" in outputs)
        self.assertFalse("Template:" in outputs)

    def test_duplcate_flags(self):
        fil = self.get_file_path("good_dupFlags.json")
        prettystring = bosh.pprint(fil)
        self.assertIn("-duplicate", prettystring)
        self.assertIn("-duplicate_DUP1", prettystring)
        self.assertIn("-duplicate_DUP2", prettystring)



#!/usr/bin/env python

import os
import uuid

import simplejson as json

from boutiques.logger import raise_error
from boutiques.util.utils import loadJson


class ExportError(Exception):
    pass


class Exporter:

    def __init__(self, descriptor, identifier, sandbox=False):
        self.descriptor = descriptor
        self.identifier = identifier
        self.sandbox = sandbox

    def convert_type(self, boutiques_type, is_integer=False, is_list=False):
        if is_list:
            return "List"
        if boutiques_type == "Flag":
            return "Boolean"
        if boutiques_type == "Number":
            if is_integer:
                return "Int64"
            return "Double"
        return boutiques_type

    def convert_input_or_output(self, input_or_output, is_output):
        param = {}
        param["name"] = input_or_output.get("name")
        param["id"] = input_or_output.get("id")
        if is_output:
            param["type"] = "File"
        else:
            param["type"] = self.convert_type(
                input_or_output.get("type"),
                input_or_output.get("integer"),
                input_or_output.get("list"),
            )
        param["isOptional"] = input_or_output.get("optional") or False
        param["isReturnedValue"] = is_output
        if input_or_output.get("default-value"):
            param["defaultValue"] = input_or_output.get("default-value")
        if input_or_output.get("description"):
            param["description"] = input_or_output.get("description")
        return param

    def carmin(self, output_file):
        carmin_desc = {}
        descriptor = loadJson(self.descriptor, sandbox=self.sandbox)

        if descriptor.get("doi"):
            self.identifier = descriptor.get("doi")

        if self.identifier is None:
            raise_error(
                ExportError,
                "Descriptor must have a DOI, or "
                "identifier must be specified with --identifier.",
            )

        carmin_desc["identifier"] = self.identifier
        carmin_desc["name"] = descriptor.get("name")
        carmin_desc["version"] = descriptor.get("tool-version")
        carmin_desc["description"] = descriptor.get("description")
        carmin_desc["canExecute"] = True
        carmin_desc["parameters"] = []
        for inp in descriptor.get("inputs"):
            carmin_desc["parameters"].append(self.convert_input_or_output(inp, False))
        for output in descriptor.get("output-files"):
            carmin_desc["parameters"].append(self.convert_input_or_output(output, True))
        carmin_desc["properties"] = {}
        carmin_desc["properties"]["boutiques"] = True
        if descriptor.get("tags"):
            for prop in descriptor.get("tags").keys():
                carmin_desc["properties"][prop] = descriptor["tags"][prop]
        carmin_desc["errorCodesAndMessages"] = []
        for errors in descriptor.get("error-codes"):
            obj = {}
            obj["errorCode"] = errors["code"]
            obj["errorMessage"] = errors["description"]
            carmin_desc["errorCodesAndMessages"].append(obj)

        with open(output_file, "w") as fhandle:
            fhandle.write(json.dumps(carmin_desc, indent=4, sort_keys=True))



from .bids import validate_bids
from .bosh import *
from .creator import CreateDescriptor
from .evaluate import evaluateEngine
from .invocationSchemaHandler import generateInvocationSchema
from .localExec import LocalExecutor
from .prettyprint import PrettyPrinter
from .publisher import Publisher
from .validator import validate_descriptor

__all__ = [
    "localExec",
    "invocationSchemaHandler",
    "creator",
    "validator",
    "bids",
    "publisher",
    "evaluate",
    "example",
    "prettyprint",
    "bosh",
]



#!/usr/bin/env python

import argparse
import os
import os.path as op
import re
import subprocess
import sys
import tempfile
from argparse import ArgumentParser

import simplejson as json
from jsonschema import ValidationError, validate

from boutiques import __file__ as bfile
from boutiques.logger import print_info, print_warning, raise_error
from boutiques.util.utils import camelCaseInputIds, customSortDescriptorByKey, loadJson


# An exception class specific to creating descriptors.
class CreatorError(ValidationError):
    pass


class CreateDescriptor:
    def __init__(self, parser=None, **kwargs):
        template = op.join(op.split(bfile)[0], "templates", "basic.json")
        with open(template) as f:
            self.descriptor = json.load(f)

        if kwargs.get("docker_image"):
            self.parse_docker(
                self.descriptor,
                kwargs.get("docker_image"),
                kwargs.get("use_singularity"),
            )

        self.sp_count = 0
        if parser is not None:
            self.parser = parser
            self.descriptor["inputs"] = []
            self.descriptor["tags"] = kwargs.get("tags") or {}
            del self.descriptor["groups"]
            del self.descriptor["output-files"]
            del self.descriptor["container-image"]
            del self.descriptor["error-codes"]
            if type(parser) is not argparse.ArgumentParser:
                raise_error(CreatorError, "Invalid argument parser")
            self.parseParser(**kwargs)

        self.camelCase = kwargs.get("camel_case")
        if self.camelCase:
            self.descriptor = camelCaseInputIds(self.descriptor)

        if kwargs.get("cl_template"):
            self.generateInputsFromTemplate(kwargs.get("cl_template"))

    def save(self, filename):
        with open(filename, "w") as f:
            f.write(json.dumps(customSortDescriptorByKey(self.descriptor), indent=4))

    def createInvocation(self, arguments):
        argdict = vars(arguments)
        argdict = {k: v for k, v in argdict.items() if v is not None}
        return argdict

    def parse_docker(self, descriptor, docker_image_name, use_singularity):
        cont_image = {}

        # Basic image config
        if use_singularity:
            cont_image["type"] = "singularity"
            cont_image["index"] = "docker://"
            cont_image["image"] = docker_image_name
        else:
            cont_image["type"] = "docker"
            cont_image["image"] = docker_image_name

        descriptor["container-image"] = cont_image

        # If Docker isn't installed, that's all we can do!
        if subprocess.Popen("type docker", shell=True).wait():
            return

        # If Docker is here, let's fetch metadata from the image
        # Properties found in the image metadata
        ((stdout, stderr), returncode) = self.executor(
            "docker pull " + docker_image_name
        )
        if returncode:
            raise_error(
                CreatorError,
                "Cannot pull Docker image {}: {} "
                "{} {}".format(docker_image_name, stdout, os.linesep, stderr),
            )
        ((stdout, stderr), returncode) = self.executor(
            "docker inspect " + docker_image_name
        )
        if returncode:
            raise_error(
                CreatorError,
                "Cannot inspect Docker image {}: {} "
                "{} {}".format(docker_image_name, stdout, os.linesep, stderr),
            )
        image_attrs = json.loads(stdout.decode("utf-8"))[0]
        if image_attrs.get("ContainerConfig"):
            container_config = image_attrs["ContainerConfig"]
            entrypoint = container_config.get("Entrypoint")
            if entrypoint:
                cont_image["entrypoint"] = True
                tokens = descriptor["command-line"].split(" ")
                # Replace the first token in the command line template,
                # presumably the executable, by the entry point
                descriptor["command-line"] = " ".join(entrypoint + tokens[1:])
                descriptor["name"] = entrypoint[0]
            workingDir = container_config.get("WorkingDir")
            if workingDir:
                raise_error(
                    CreatorError,
                    "The container image has a working "
                    "dir, this is currently not supported.",
                )
        if image_attrs.get("Author"):
            descriptor["author"] = image_attrs.get("Author")
        if image_attrs.get("RepoTags"):
            descriptor["tool-version"] = " ".join(image_attrs.get("RepoTags"))
        if image_attrs.get("Comment"):
            descriptor["description"] = image_attrs.get("Comment")

    def parseParser(self, **kwargs):
        self.descriptor["command-line"] = kwargs.get("execname")
        for act in self.parser._actions:
            tmp = self.parseAction(act, **kwargs)
            if bool(tmp):
                self.descriptor["inputs"] += [tmp]

    def parseAction(self, action, **kwargs):
        # Case 1: input is a help flag
        # Desired outcome: we skip it
        if type(action) is argparse._HelpAction:
            if kwargs.get("verbose"):
                print_info("_HelpAction: Skipping")
            # If this action belongs to a subparser, return a flag alongside
            # the empty object, indicating it is not required
            if kwargs.get("subaction"):
                return {}, False
            return {}

        # Case 2: input is a subparser
        # Desired outcome: we add the subparser and options, and an input for
        # each of the subparser options
        elif type(action) is argparse._SubParsersAction and not kwargs.get("addParser"):
            if kwargs.get("verbose"):
                print_info("_SubParsersAction: Interpreting & Adding")

            # First, add the subparser itself as an input.
            subparser = self.parseAction(action, addParser=True)
            subparser["value-requires"] = {}
            inpts = {}
            # For every option specified by the subparser...
            for act in subparser["value-choices"]:
                inpts[act] = []
                subparser["value-requires"][act] = []
                # ... And for each choice specified by each subparser...
                for subact in action.choices[act]._actions:

                    # Process the action, and record its "required" status
                    tmpinput, reqd = self.parseAction(subact, subaction=True, **kwargs)

                    # If it's not empty, add it to an inputs dictionaryi, and
                    # add the input to the descriptor.
                    if tmpinput != {}:
                        inpts[act] += [tmpinput["id"]]
                        # If the input was required by the subparser, record it
                        if reqd:
                            subparser["value-requires"][act] += [tmpinput["id"]]
                        self.descriptor["inputs"] += [tmpinput]

            # Once all subparsers are processed, identify which inputs need to
            # be disabled by which subparsers.
            inpt_ids = {inp for iact in inpts for inp in inpts[iact]}
            subparser["value-disables"] = {}
            for act in subparser["value-choices"]:
                # Add all IDs created by the subparser that do not also belong
                # to the current selection to the disabled list.
                subparser["value-disables"][act] = [
                    ckey for ckey in inpt_ids if ckey not in inpts[act]
                ]
            return subparser

        # Case 3: input is a regular input
        # Desired outcome: we add it, unless it's already been added
        else:
            if kwargs.get("verbose"):
                actstring = str(type(action))
                actstring = actstring.split("'")[1].split(".")[-1]
                print_info(f"{actstring}: Adding")
            actdict = vars(action)
            if action.dest == "==SUPPRESS==":
                adest = f"subparser_{self.sp_count}"
                if kwargs.get("verbose"):
                    print_warning(
                        "Subparser has no destination set, "
                        "invocation parsing may not work as "
                        "expected. This can be fixed by adding "
                        "\"dest='mysubparser'\" to subparser "
                        "creation."
                    )
                self.sp_count += 1
            else:
                adest = action.dest

            # If an input already exists with this ID, don't re-add it
            if any(adest == it["id"] for it in self.descriptor["inputs"]):
                if kwargs.get("verbose"):
                    print_info(
                        f"Duplicate: Argument won't be added multiple times ({adest})"
                    )
                # If this action belongs to a subparser return a flag alongside
                # the empty object, indicating it is not required
                if kwargs.get("subaction"):
                    return {}, False
                return {}

            # If no argument exists yet by this name, process and add it.
            # First, by setting some reasonable defaults or obvious values,
            # and then by updating others.
            newinput = {
                "id": adest,
                "name": adest,
                "description": action.help,
                "optional": kwargs.get("subaction") or not action.required,
                "type": "String",
                "value-key": f"[{adest.upper().strip('[]')}]",
            }

            if action.type is not None:
                if action.type in [int, float]:
                    newinput["type"] = "Number"
                if action.type == list:
                    newinput["list"] = True

            if action.nargs is not None:
                if (isinstance(action.nargs, str) and action.nargs == "+") or (
                    isinstance(action.nargs, int) and action.nargs > 1
                ):
                    newinput["list"] = True

            if action.default:
                newinput["default-value"] = action.default

            if action.choices:
                try:
                    # Subparsers have choices in the form of OrderedDicts...
                    newinput["value-choices"] = list(action.choices.keys())
                except AttributeError as e:
                    # ... but "choice" inputs have them in the form a list.
                    newinput["value-choices"] = action.choices

            if len(action.option_strings):
                newinput["command-line-flag"] = action.option_strings[0]

            if type(action) is argparse._StoreTrueAction:
                newinput["type"] = "Flag"

            self.descriptor["command-line"] += f" [{adest.upper().strip('[]')}]"
            # If this action belongs to a subparser, return a flag along
            # with the object, indicating its required/not required status.
            if kwargs.get("subaction"):
                return newinput, action.required
            return newinput

    def executor(self, command):
        try:
            process = subprocess.Popen(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except OSError as e:
            sys.stderr.write("OS Error during attempted execution!")
            raise e
        except ValueError as e:
            sys.stderr.write("Input Value Error during attempted execution!")
            raise e
        else:
            return process.communicate(), process.returncode

    def generateInputsFromTemplate(self, cl_template):
        def _createIdFromValueKey(vk):
            vk = re.sub(r"\[(\w+)\]", r"\1", vk).lower()
            vk_words = vk.split("_")
            return vk_words[0] + "".join(x.title() for x in vk_words[1:])

        def _createNameFromId(id):
            name = [c for c in id]
            for idx, c in enumerate(name):
                if not c.isalnum():
                    name[idx] = " "
                if c.isupper():
                    name[idx] = " " + c.lower()
            # Join char list into string and concatenate white spaces
            return " ".join("".join(name).split()).title()

        if os.path.isfile(cl_template):
            template_descriptor = loadJson(cl_template)
            cl_template = template_descriptor["command-line"]

        self.descriptor["command-line"] = cl_template
        # Clear descriptor inputs and repopulate
        self.descriptor["inputs"] = []
        for vk in re.findall(r"\[\w+\]", cl_template):
            generated_id = _createIdFromValueKey(vk)
            self.descriptor["inputs"].append(
                {
                    "name": _createNameFromId(generated_id),
                    "id": generated_id,
                    "type": "String",
                    "value-key": vk,
                }
            )



#!/usr/bin/env python

import os
import shutil
from unittest import mock

import pytest

import boutiques
from boutiques.bosh import bosh
from boutiques.nexusHelper import NexusError
from boutiques.tests.BaseTest import BaseTest
from boutiques.tests.boutiques_mocks import (
    mock_empty_function,
    mock_get_data_cache,
    mock_get_data_cache_file,
    mock_get_empty_nexus_credentials,
    mock_get_invalid_nexus_endpoint,
    mock_get_publish_bulk,
    mock_get_publish_single,
    mock_post_publish_bulk,
    mock_post_publish_single,
)

ZENODO_SANDBOX_TOKEN = "fake-token-123"


class TestDataHandler(BaseTest):
    @pytest.fixture(autouse=True)
    def reset(self):
        # Reset before each test
        if os.path.isdir(mock_get_data_cache()):
            shutil.rmtree(mock_get_data_cache())
        src = self.get_file_path("data_collection/input_dir/test-data-cache")
        shutil.copytree(src, mock_get_data_cache())

    # Captures the stdout and stderr during test execution
    # and returns them as a tuple in readouterr()
    @pytest.fixture(autouse=True)
    def capture_st(self, capfd):
        self.capfd = capfd

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    def test_inspect(self, mock_dir):
        boutiques.data("inspect")
        out, _ = self.capfd.readouterr()
        self.assertIn("There are 3 unpublished records in the cache", out)

        boutiques.data("inspect", "-e")
        out, _ = self.capfd.readouterr()
        self.assertIn("summary", out)
        self.assertIn("descriptor-doi", out)

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    def test_delete(self, mock_dir):
        with self.assertRaises(ValueError) as e:
            bosh(["data", "delete", "-f", "bad-filename", "--no-int"])
        self.assertIn(
            "File bad-filename does not exist in the data cache",
            str(e.exception),
        )

        fl1_path = os.path.join(mock_get_data_cache(), "tool1_123.json")
        fl2_path = os.path.join(mock_get_data_cache(), "tool2_123.json")
        bosh(["data", "delete", "-f", "tool1_123.json", "-y"])
        self.assertFalse(os.path.isfile(fl1_path))
        bosh(["data", "delete", "-f", fl2_path, "-y"])
        self.assertFalse(os.path.isfile(fl2_path))
        bosh(["data", "delete", "-y"])
        self.assertEqual(len(os.listdir(mock_get_data_cache())), 0)

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    @mock.patch("requests.get", side_effect=mock_get_publish_single())
    @mock.patch("requests.post", side_effect=mock_post_publish_single())
    def test_publish_single(self, mock_dir, mock_get, mock_post):
        # Publish a record that does not exist
        with self.assertRaises(ValueError) as e:
            bosh(["data", "publish", "-f", "bad-filename", "--sandbox", "-y"])
        self.assertIn(
            "File bad-filename does not exist in the data cache",
            str(e.exception),
        )

        # Publish a record with a correct descriptor-doi
        bosh(
            [
                "data",
                "publish",
                "-f",
                "tool1_123.json",
                "-y",
                "--sandbox",
                "--zenodo-token",
                ZENODO_SANDBOX_TOKEN,
            ]
        )
        self.assertFalse(
            os.path.isfile(os.path.join(mock_get_data_cache(), "tool1-123"))
        )

        # Publish a record without a descriptor-doi
        bosh(
            [
                "data",
                "publish",
                "-f",
                "tool2_123.json",
                "-y",
                "--sandbox",
                "--zenodo-token",
                ZENODO_SANDBOX_TOKEN,
            ]
        )
        self.assertTrue(
            os.path.isfile(os.path.join(mock_get_data_cache(), "tool2_123.json"))
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(mock_get_data_cache(), "descriptor_tool2_123.json")
            )
        )

        # Publish a record without a descriptor-doi but descriptor is published
        bosh(
            [
                "data",
                "publish",
                "-f",
                "tool3_123.json",
                "-y",
                "--sandbox",
                "--zenodo-token",
                ZENODO_SANDBOX_TOKEN,
            ]
        )
        self.assertFalse(
            os.path.isfile(os.path.join(mock_get_data_cache(), "tool3-123"))
        )
        self.assertFalse(
            os.path.isfile(os.path.join(mock_get_data_cache(), "descriptor-tool3-123"))
        )

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    @mock.patch("requests.get", side_effect=mock_get_publish_bulk())
    @mock.patch("requests.post", side_effect=mock_post_publish_bulk())
    def test_publish_bulk(self, mock_dir, mock_get, mock_post):
        bosh(
            [
                "data",
                "publish",
                "-y",
                "--sandbox",
                "--zenodo-token",
                ZENODO_SANDBOX_TOKEN,
            ]
        )
        self.assertEqual(len(os.listdir(os.path.join(mock_get_data_cache()))), 2)

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    @mock.patch("requests.get", side_effect=mock_get_publish_single())
    @mock.patch("requests.post", side_effect=mock_post_publish_single())
    def test_publish_individual(self, mock_dir, mock_get, mock_post):
        bosh(
            [
                "data",
                "publish",
                "-y",
                "--individual",
                "--sandbox",
                "--zenodo-token",
                ZENODO_SANDBOX_TOKEN,
            ]
        )
        self.assertEqual(len(os.listdir(os.path.join(mock_get_data_cache()))), 2)

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    @mock.patch(
        "boutiques.nexusHelper.NexusHelper.read_credentials",
        return_value=mock_get_empty_nexus_credentials(),
    )
    def test_publish_nexus_no_token_fail(self, mock_dir, mock_cred):
        with self.assertRaises(NexusError) as e:
            bosh(["data", "publish", "-y", "--nexus", "--sandbox"])
        self.assertIn("Cannot find Nexus credentials.", str(e.exception))

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    @mock.patch(
        "boutiques.nexusHelper.NexusHelper.read_credentials",
        return_value=mock_get_empty_nexus_credentials(),
    )
    def test_publish_nexus_no_organization_fail(self, mock_dir, mock_cred):
        with self.assertRaises(NexusError) as e:
            bosh(
                [
                    "data",
                    "publish",
                    "-y",
                    "--nexus",
                    "--sandbox",
                    "--nexus-token",
                    "hAaW2wSBZMskxpfigTYHcuDrCPWr2",
                ]
            )
        self.assertIn("Cannot find Nexus organization.", str(e.exception))

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    @mock.patch(
        "boutiques.nexusHelper.NexusHelper.read_credentials",
        return_value=mock_get_empty_nexus_credentials(),
    )
    def test_publish_nexus_no_project_fail(self, mock_dir, mock_cred):
        with self.assertRaises(NexusError) as e:
            bosh(
                [
                    "data",
                    "publish",
                    "-y",
                    "--nexus",
                    "--sandbox",
                    "--nexus-token",
                    "hAaW2wSBZMskxpfigTYHcuDrCPWr2",
                    "--nexus-org",
                    "boutiques",
                ]
            )
        self.assertIn("Cannot find Nexus project.", str(e.exception))

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    @mock.patch(
        "boutiques.nexusHelper.NexusHelper.get_nexus_endpoint",
        return_value=mock_get_invalid_nexus_endpoint(),
    )
    def test_publish_nexus_invalid_endpoint(self, mock_dir, mock_endpoint):
        with self.assertRaises(NexusError) as e:
            bosh(
                [
                    "data",
                    "publish",
                    "-y",
                    "--nexus",
                    "--sandbox",
                    "--nexus-token",
                    "hAaW2wSBZMskxpfigTYHcuDrCPWr2",
                    "--nexus-org",
                    "boutiques",
                    "--nexus-project",
                    "test",
                ]
            )
        self.assertIn("Cannot access Nexus endpoint", str(e.exception))

    @pytest.mark.xfail(
        reason="Nexus sandbox API is not available for testing currently"
    )
    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    def test_publish_nexus_invalid_token(self, mock_dir):
        with self.assertRaises(NexusError) as e:
            bosh(
                [
                    "data",
                    "publish",
                    "-y",
                    "--nexus",
                    "--sandbox",
                    "--nexus-token",
                    "INVALIDTOKEN",
                    "--nexus-org",
                    "boutiques",
                    "--nexus-project",
                    "test",
                ]
            )
        self.assertIn(
            "Cannot authenticate to Nexus API, check " "your access token",
            str(e.exception),
        )

    @mock.patch(
        "boutiques.dataHandler.getDataCacheDir",
        return_value=mock_get_data_cache(),
    )
    @mock.patch(
        "boutiques.nexusHelper.NexusHelper.get_config_file",
        return_value=mock_get_data_cache_file(),
    )
    @mock.patch(
        "boutiques.nexusHelper.NexusHelper.nexus_test_api",
        return_value=mock_empty_function(),
    )
    @mock.patch("nexussdk.files.create", return_value=mock_empty_function())
    def test_publish_nexus_success(
        self, mock_create_file, mock_fetch_project, mock_save_inputs, mock_dir
    ):
        try:
            bosh(
                [
                    "data",
                    "publish",
                    "-y",
                    "--nexus",
                    "--sandbox",
                    "--nexus-token",
                    "hAaW2wSBZMskxpfigTYHcuDrCPWr2",
                    "--nexus-org",
                    "test",
                    "--nexus-project",
                    "test",
                    "-v",
                ]
            )
        except Exception as e:
            self.fail("Unexpected exception raised: " + str(e))



#!/usr/bin/env python

import pytest

from boutiques.bids import validate_bids
from boutiques.bosh import bosh
from boutiques.tests.BaseTest import BaseTest
from boutiques.validator import DescriptorValidationError


class TestBIDS(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("bids")

    def test_bids_good(self):
        fil = self.get_file_path("bids_good.json")
        self.assertFalse(bosh(["validate", fil, "-b"]))

    def test_bids_bad1(self):
        fil = self.get_file_path("bids_bad1.json")
        self.assertRaises(DescriptorValidationError, bosh, ["validate", fil, "-b"])

    def test_bids_bad2(self):
        fil = self.get_file_path("bids_bad2.json")
        self.assertRaises(DescriptorValidationError, bosh, ["validate", fil, "-b"])

    def test_bids_invalid(self):
        import simplejson as json

        fil = self.get_file_path("bids_bad2.json")
        descriptor = json.load(open(fil))
        self.assertRaises(DescriptorValidationError, validate_bids, descriptor, False)



"""
Naval Fate.

Usage:
  naval_fate ship new <name>...
  naval_fate ship <name> move <x> <y> [--speed=<kn>]
  naval_fate ship shoot <x> <y>
  naval_fate mine (set|remove) <x> <y> [--moored|--drifting]
  naval_fate -h | --help
  naval_fate --version

Options:
  -h --help     Show this screen.
  -v --version     Show version.
  --speed=<kn>  Speed in knots [default: 10].
  --moored      Moored (anchored) mine.
  --drifting    Drifting mine.
  """



#!/usr/bin/env python

import os
import subprocess
from pathlib import Path

import pytest

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.tests.BaseTest import BaseTest


class TestExample3(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup(
            os.path.join(os.path.dirname(bfile), "schema", "examples", "example3")
        )

    def test_example3_validate(self):
        self.assertIsNone(bosh.validate(self.get_file_path("example3.json")))

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example3_exec(self):
        Path(self.test_temp, "FileValue.txt").touch()
        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.get_file_path("example3.json"),
                self.get_file_path("invocation.json"),
                "--skip-data-collection",
            ),
            aditional_assertions=self.assert_no_output,
        )

    def test_example3_filepathrenaming(self):
        self.assertEqual(
            bosh.evaluate(
                self.get_file_path("example3.json"),
                self.get_file_path("invocation.json"),
                "output-files/",
            ),
            {"logfile": "./test_temp/log-FileValue.txt"},
        )



#!/usr/bin/env python
# -*- coding: latin-1 -*-

import argparse
import os
import sys


def file_exists(parser, file_name):
    if not os.path.exists(file_name):
        parser.error("File not found: %s" % file_name)
    return file_name


def is_valid_enum_value(parser, value):
    if not value in ["val1", "val2", "val3"]:
        parser.error("Invalid enum value: %s" % value)
    return value


def main(args=None):
    parser = argparse.ArgumentParser(
        description="A tool to test Boutiques",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-c",
        "--config_file",
        type=lambda x: file_exists(parser, x),
        help="A config file with a number.",
    )
    parser.add_argument("-i", "--string_input", nargs="+", help="One or more strings.")
    parser.add_argument(
        "-l", "--int_input", nargs="+", help="One or more ints.", type=int
    )
    parser.add_argument(
        "-t",
        "--file_list",
        nargs="+",
        help="One or more files.",
        type=lambda x: file_exists(parser, x),
    )
    parser.add_argument(
        "file",
        metavar="file_input",
        type=lambda x: file_exists(parser, x),
        nargs=1,
        help="Any existing file.",
    )
    parser.add_argument("-f", "--flag", action="store_true", help="A flag.")
    parser.add_argument(
        "-n",
        "--number",
        type=int,
        help="A number. Cannot be used when enum_input is used",
    )
    parser.add_argument("-s", "--string", help="A string")
    parser.add_argument(
        "-e",
        "--enum_input",
        type=lambda x: is_valid_enum_value(parser, x),
        nargs=1,
        help='A string in {"val1", "val2", "val3"}.',
    )
    parser.add_argument("output_file", help="The output file.")

    results = parser.parse_args() if args is None else parser.parse_args(args)

    if results.number and results.enum_input:
        sys.stderr.write("error: number and enum_input are mutually exclusive.\n")
        sys.exit(1)

    if (not results.number) and (not results.enum_input):
        sys.stderr.write("error: one of number or enum_input is required.\n")
        sys.exit(1)

    var = os.getenv("ENVAR")
    if var != "theValue":
        sys.stderr.write(
            "error: ENVAR environment variable must be set to 'theValue'.\n"
        )
        sys.exit(1)

    homeDir = os.getenv("HOME")
    if homeDir != os.getenv("PWD"):
        sys.stderr.write(
            "error: HOME environment variable must be set to the current"
            " working directory.\n"
        )
        sys.exit(1)

    with open(results.config_file, "r") as co:
        config_string = co.read()
        if (
            config_string
            != "# This is a demo configuration file\nnumInput=4\nstrInput='coin;plop'"
        ):
            sys.stderr.write(
                "error: invalid configuration file:\n %s\n" % config_string
            )
            sys.exit(1)

    sys.stdout.write("Ça c'est stdout")
    sys.stderr.write("This is stdérr")
    with open(results.output_file, "w") as f:
        f.write("File content")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os

import pytest

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.tests.BaseTest import BaseTest


class TestExample2(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup(
            os.path.join(os.path.dirname(bfile), "schema", "examples", "example2")
        )

    def test_example2_validate(self):
        self.assertIsNone(bosh.validate(self.get_file_path("example2.json")))

    def test_example2_no_exec(self):
        self.assert_successful_return(
            bosh.execute(
                "simulate",
                self.get_file_path("example2.json"),
                "-i",
                self.get_file_path("invocation.json"),
            ),
            aditional_assertions=self.assert_only_stdout,
        )

    def test_example2_exec(self):
        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.get_file_path("example2.json"),
                self.get_file_path("invocation.json"),
                "--skip-data-collection",
            ),
            aditional_assertions=self.assert_no_output,
        )

        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.get_file_path("example2.json"),
                "-x",
                self.get_file_path("invocation.json"),
                "--skip-data-collection",
            ),
            aditional_assertions=self.assert_no_output,
        )

    def test_example2_no_exec_random(self):
        self.assertFalse(
            bosh.execute("simulate", self.get_file_path("example2.json")).exit_code
        )



#!/usr/bin/env python

import os
import subprocess

import pytest
import simplejson as json

from boutiques.tests.BaseTest import BaseTest


class TestOutputFiles(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("output_files")

    @pytest.fixture(autouse=True)
    def clean_up(self):
        yield
        if os.path.exists(self.get_file_path("get_test_desc.json")):
            os.remove(self.get_file_path("get_test_desc.json"))

    def test_output_conditional_path_template_validity(self):
        test_desc_path = self.get_file_path("test_fixANDcond_output_desc.json")

        # Validate descriptor with output-files containing both conditional
        # and fixed path-templates objects
        command = "bosh validate " + test_desc_path
        validate = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        output = validate.stdout.read()
        self.assertEqual(b"OK\n", output)

    def test_conditional_path_template_mutual_exclusivity(self):
        base_desc_path = self.get_file_path("test_fixANDcond_output_desc.json")
        test_desc_path = self.get_file_path("test_desc.json")

        template_desc_json = {}
        with open(base_desc_path) as base_desc:
            template_desc_json = json.load(base_desc)

        test_json = {
            k: template_desc_json[k] for k in template_desc_json if k != "output-files"
        }
        output_list = template_desc_json["output-files"]

        # Test descriptor with only path template output file
        test_json["output-files"] = [
            out for out in output_list if "path-template" in out
        ]
        with open(test_desc_path, "w+") as test_desc:
            test_desc.write(json.dumps(test_json))
        command = "bosh validate " + test_desc_path
        validate_only_PT = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        process_output = validate_only_PT.stdout.read()
        self.assertEqual(b"OK\n", process_output)

        # Test descriptor with only conditional path template output file
        test_json["output-files"] = [
            out for out in output_list if "path-template" not in out
        ]
        with open(test_desc_path, "w+") as test_desc:
            test_desc.write(json.dumps(test_json))
        validate_only_CPT = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE
        )
        process_output = validate_only_CPT.stdout.read()
        self.assertEqual(b"OK\n", process_output)

        # Test descriptor with output-files object containing both \
        # conditional and fixed path-templates
        test_json["output-files"][0]["path-template"] = "[PARAM1]/out1.txt"
        with open(test_desc_path, "w+") as test_desc:
            test_desc.write(json.dumps(test_json))
        validate_fixANDCond = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE
        )
        process_output = validate_fixANDCond.stdout.read()
        self.assertIn("[ ERROR ]", process_output.decode())
        self.assertIn("schema", process_output.decode())

    def test_conditional_path_template_optionality(self):
        base_desc_path = self.get_file_path("test_fixANDcond_output_desc.json")
        test_desc_path = self.get_file_path("test_desc.json")

        template_desc_json = {}
        with open(base_desc_path) as base_desc:
            template_desc_json = json.load(base_desc)

        test_json = {
            k: template_desc_json[k] for k in template_desc_json if k != "output-files"
        }
        output_list = template_desc_json["output-files"]

        # Test non-optional conditional-path-template with missing default
        test_json["output-files"] = [
            out for out in output_list if "path-template" not in out
        ]
        test_json["output-files"][0]["conditional-path-template"].remove(
            {"default": "[PARAM2]_default.txt"}
        )
        with open(test_desc_path, "w+") as test_desc:
            test_desc.write(json.dumps(test_json))
        command = "bosh validate " + test_desc_path
        validate_missing_default = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE
        )
        process_output = validate_missing_default.stdout.read()
        self.assertIn(
            'OutputError: "out2". Non-optional output-file with '
            "conditional-path-template must contain "
            '"default" path-template.',
            process_output.decode(),
        )

        # Test duplicate default condition in CPT
        test_json["output-files"][0]["conditional-path-template"].extend(
            [
                {"default": "[PARAM2]_default.txt"},
                {"default": "[PARAM1]_default.txt"},
            ]
        )
        with open(test_desc_path, "w+") as test_desc:
            test_desc.write(json.dumps(test_json))
        validate_duplicate_default = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE
        )
        process_output = validate_duplicate_default.stdout.read()
        self.assertIn(
            'OutputError: "out2". Only one "default" condition '
            "is permitted in a conditional-path-template.",
            process_output.decode(),
        )

    def test_conditional_path_template_expressions(self):
        base_desc_path = self.get_file_path("test_fixANDcond_output_desc.json")
        test_desc_path = self.get_file_path("test_desc.json")

        template_desc_json = {}
        with open(base_desc_path) as base_desc:
            template_desc_json = json.load(base_desc)

        test_json = {
            k: template_desc_json[k] for k in template_desc_json if k != "output-files"
        }
        output_list = template_desc_json["output-files"]

        # Test non-optional conditional-path-template with wrong ID name
        test_json["output-files"] = [
            out for out in output_list if "path-template" not in out
        ]
        test_json["output-files"][0]["conditional-path-template"].append(
            {"bob": "invalid.txt"}
        )
        with open(test_desc_path, "w+") as test_desc:
            test_desc.write(json.dumps(test_json))
        command = "bosh validate " + test_desc_path
        validate_wrong_ID = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE
        )
        process_output = validate_wrong_ID.stdout.read()
        self.assertIn(
            'OutputError: "out2" contains non-python '
            'keyword and non-ID string: "bob"',
            process_output.decode(),
        )

    def test_conditional_path_template_comparison_types(self):
        test_desc_path = self.get_file_path("test_cond_output_invalid_types.json")

        command = "bosh validate " + test_desc_path
        validate_wrong_ID = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE
        )
        process_output = validate_wrong_ID.stdout.read()
        self.assertIn(
            '"opt1 and (opt2 > "10.00")" contains ' "invalid conditional expression.",
            process_output.decode(),
        )
        self.assertIn(
            '"opt1 and opt2 and (opt1 == "False")" contains '
            "invalid conditional expression.",
            process_output.decode(),
        )
        self.assertIn(
            '"param1 < 10.0" contains ' "invalid conditional expression.",
            process_output.decode(),
        )



#!/usr/bin/env python

import subprocess

import pytest

from boutiques.tests.BaseTest import BaseTest


class TestCrashPython3(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("crash_python3")

    @pytest.mark.usefixtures("skip_if_no_apptainer")
    def test_no_container(self):
        command = "bosh exec launch --skip-data-collection " "{} {}".format(
            self.get_file_path("crash3.json"),
            self.get_file_path("crash3_invocation.json"),
        )
        print(command)
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout = process.stdout.read().decode("utf-8").strip()
        self.assertIn("Could not pull Singularity image", stdout)



import os

from boutiques import __file__ as bfile


class MockHttpResponse:
    def __init__(self, status_code, mock_json={}):
        self.status_code = status_code
        self.mock_json = mock_json

    def json(self):
        return self.mock_json


ZENODO_RECORD = 129904
ZENODO_FILE = (
    f"https://sandbox.zenodo.org/records/{ZENODO_RECORD}/files/example1_docker.json"
)


class MockZenodoRecord:
    def __init__(
        self,
        id,
        title,
        description="",
        filename="",
        downloads=1,
        keywords=[],
        is_last_version=True,
    ):
        self.id = id
        self.title = title
        self.filename = filename
        self.downloads = downloads
        self.description = description
        self.keywords = keywords
        self.is_last_version = is_last_version

    def reset(
        self,
        id=ZENODO_RECORD,
        title="Example Boutiques Tool",
        description="",
        filename=ZENODO_FILE,
        downloads=1,
        keywords=[],
        is_last_version=True,
    ):
        self.id = id
        self.title = title
        self.filename = filename
        self.downloads = downloads
        self.description = description
        self.keywords = keywords
        self.is_last_version = is_last_version


# Mock object representing the current version of Example Boutiques Tool
# on Zenodo
example_boutiques_tool = MockZenodoRecord(
    ZENODO_RECORD,
    "Example Boutiques Tool",
    filename=ZENODO_FILE,
)


def mock_get():
    return mock_zenodo_search(
        [
            MockZenodoRecord(
                ZENODO_RECORD,
                "Example Boutiques Tool",
                filename=ZENODO_FILE,
            )
        ]
    )


def mock_zenodo_test_api(*args, **kwargs):
    return MockHttpResponse(200)


def mock_zenodo_upload_descriptor():
    return MockHttpResponse(201)


def mock_zenodo_deposit(mock_zid):
    mock_json = {"id": mock_zid}
    return MockHttpResponse(201, mock_json)


def mock_zenodo_publish(mock_zid):
    mock_json = {"doi": f"10.5281/zenodo.{mock_zid}"}
    return MockHttpResponse(202, mock_json)


def mock_zenodo_delete_files(*args, **kwargs):
    return MockHttpResponse(204)


def mock_zenodo_test_api_fail():
    return MockHttpResponse(403)


def mock_zenodo_no_permission():
    return MockHttpResponse(403)


def mock_get_invalid_nexus_endpoint():
    return "https://invalid.nexus.endpoint/v1"


def mock_get_empty_nexus_credentials():
    return {}


def mock_empty_function():
    return


def get_zenodo_record(record, include_version=True):
    record = {
        "doi": f"10.5281/zenodo.{record.id}",
        "files": [{"links": {"self": record.filename}}],
        "id": record.id,
        "metadata": {
            "creators": [{"name": "Test author"}],
            "description": record.description,
            "publication_date": "2020-04-15",
            "relations": {
                "version": [
                    {
                        "count": 4,
                        "index": 3,
                        "is_last": record.is_last_version,
                        "last_child": {"pid_value": "33333"},
                    }
                ]
            },
            "doi": f"10.5281/zenodo.{record.id}",
            "keywords": ["schema-version:0.5", "docker"],
            "title": record.title,
            "version": "0.0.1",
        },
        "stats": {"version_downloads": record.downloads},
    }
    if not include_version:
        del record["metadata"]["version"]
    return record


def mock_get_data_cache():
    return os.path.join(
        os.path.split(os.path.split(bfile)[0])[0],
        "test_temp",
        "test-data-cache",
    )


def mock_zenodo_search(mock_records, include_version=True):
    mock_results = []
    for record in mock_records:
        mock_results.append(get_zenodo_record(record, include_version))
    mock_json = {"hits": {"hits": mock_results, "total": len(mock_results)}}
    return MockHttpResponse(200, mock_json)


def mock_get_data_cache_file():
    return os.path.join(
        os.path.split(os.path.split(bfile)[0])[0],
        "test_temp",
        "test-data-cache",
        "nexus",
    )


def mock_get_publish_single():
    return [
        mock_zenodo_test_api_fail(),
        mock_zenodo_test_api(),
        mock_zenodo_test_api_fail(),
        mock_zenodo_test_api(),
        mock_zenodo_test_api_fail(),
        mock_zenodo_test_api(),
        mock_zenodo_test_api_fail(),
        mock_zenodo_test_api(),
    ]


def mock_post_publish_single():
    return [
        mock_zenodo_deposit(1234567),
        mock_zenodo_upload_descriptor(),
        mock_zenodo_publish(1234567),
        mock_zenodo_deposit(1234567),
        mock_zenodo_upload_descriptor(),
        mock_zenodo_publish(1234567),
    ]


def mock_get_publish_bulk():
    return [mock_zenodo_test_api_fail(), mock_zenodo_test_api()]


def mock_post_publish_bulk():
    return [
        mock_zenodo_deposit(1234567),
        mock_zenodo_upload_descriptor(),
        mock_zenodo_upload_descriptor(),
        mock_zenodo_publish(1234567),
    ]



#!/usr/bin/env python

import os
import re
import subprocess
from unittest import mock

import pytest
import simplejson as json
from boutiques_mocks import (
    MockZenodoRecord,
    example_boutiques_tool,
    mock_get,
    mock_zenodo_search,
)

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.localExec import ExecutorError
from boutiques.tests.BaseTest import BaseTest


class TestSimulate(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("simulate")

    def test_success(self):
        self.assertFalse(bosh.execute("simulate", self.example1_descriptor).exit_code)

        self.assertFalse(
            bosh.execute(
                "simulate",
                self.example1_descriptor,
                self.get_file_path("good_nooutputs.json"),
            ).exit_code
        )

    def test_success_desc_as_json_obj(self):
        desc_json = open(self.example1_descriptor).read()
        self.assertFalse(bosh.execute("simulate", desc_json).exit_code)

    @mock.patch("requests.get", return_value=mock_get())
    def test_success_desc_from_zenodo(self, mock_get):
        self.assertFalse(
            bosh.execute(
                "simulate", "zenodo." + str(example_boutiques_tool.id)
            ).exit_code
        )

    def test_success_default_values(self):
        self.assertFalse(
            bosh.execute(
                "simulate",
                self.example1_descriptor,
                "-i",
                self.get_file_path("inv_no_defaults.json"),
            ).exit_code
        )

    @mock.patch("random.uniform")
    def test_number_bounds(self, mock_random):
        desc_json = open(self.example1_descriptor).read()
        test_json = json.loads(desc_json)
        del test_json["groups"]
        # Have to tweak flag_input because it disables number_input
        target_input = [i for i in test_json["inputs"] if i["id"] == "flag_input"][0]
        del target_input["disables-inputs"]

        # Make number_input mandatory
        target_input = [i for i in test_json["inputs"] if i["id"] == "num_input"][0]
        target_input["optional"] = False
        target_input["exclusive-minimum"] = False

        # Test inclusive lower bound
        mock_random.return_value = -0.001
        self.assertRaises(
            ExecutorError,
            bosh.execute,
            ("simulate", json.dumps(test_json), "-j"),
        )
        mock_random.return_value = 0
        ret = bosh.bosh(args=["example", json.dumps(test_json)])
        self.assertIsInstance(json.loads(ret), dict)

        # Test exclusive lower bound
        target_input["exclusive-minimum"] = True
        self.assertRaises(
            ExecutorError,
            bosh.execute,
            ("simulate", json.dumps(test_json), "-j"),
        )
        mock_random.return_value = 0.001
        ret = bosh.bosh(args=["example", json.dumps(test_json)])
        self.assertIsInstance(json.loads(ret), dict)

        # Test inclusive upper bound
        target_input["exclusive-maximum"] = False
        mock_random.return_value = 1.001
        self.assertRaises(
            ExecutorError,
            bosh.execute,
            ("simulate", json.dumps(test_json), "-j"),
        )
        mock_random.return_value = 1
        ret = bosh.bosh(args=["example", json.dumps(test_json)])
        self.assertIsInstance(json.loads(ret), dict)

        # Test exclusive upper bound
        target_input["exclusive-maximum"] = True
        self.assertRaises(
            ExecutorError,
            bosh.execute,
            ("simulate", json.dumps(test_json), "-j"),
        )
        mock_random.return_value = 0.999
        ret = bosh.bosh(args=["example", json.dumps(test_json)])
        self.assertIsInstance(json.loads(ret), dict)

    def test_success_json(self):
        desc_json = open(self.example1_descriptor).read()
        ret = bosh.execute("simulate", desc_json, "-j").stdout
        self.assertIsInstance(json.loads(ret), dict)
        ret = bosh.bosh(args=["example", desc_json])
        self.assertIsInstance(json.loads(ret), dict)

    def test_failing_bad_descriptor_invo_combos(self):
        self.setup("example1")
        self.assertRaises(
            ExecutorError,
            bosh.execute,
            (
                "simulate",
                self.get_file_path("fake.json"),
                "-i",
                self.get_file_path("invocation.json"),
            ),
        )

        self.assertRaises(
            ExecutorError,
            bosh.execute,
            (
                "simulate",
                self.example1_descriptor,
                "-i",
                self.get_file_path("fake.json"),
            ),
        )

        self.assertRaises(
            ExecutorError,
            bosh.execute,
            (
                "simulate",
                self.example1_descriptor,
                "-i",
                self.get_file_path("exampleTool1.py"),
            ),
        )

        self.assertRaises(
            ExecutorError, bosh.execute, ("simulate", self.example1_descriptor)
        )

        self.assertRaises(
            ExecutorError,
            bosh.execute,
            (
                "simulate",
                self.example1_descriptor,
                "-i",
                self.get_file_path("invocation.json"),
            ),
        )

        self.assertRaises(
            ExecutorError, bosh.execute, ("simulate", self.example1_descriptor)
        )

    def test_collapsing_whitespace_optionals(self):
        self.setup("example")
        descriptor = self.get_file_path("test_example_descriptor.json")
        command = "bosh exec simulate " + descriptor + " -c"
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        output = str(process.stdout.read())
        command = re.search(r"(test[ \da-z]+)", output).group(0)

        self.assertNotIn("  ", command)

    def test_collapsing_whitespace_requireds(self):
        self.setup("example")
        descriptor = self.get_file_path("test_example_descriptor.json")
        command = "bosh exec simulate " + descriptor
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        output = str(process.stdout.read())
        command = re.search(r"(test[ \da-z]+)", output).group(0)

        self.assertNotIn("  ", command)

    def test_consistency_withAndWithout_invoc(self):
        descriptor = self.get_file_path("test_simulate_consistency.json")
        noInvoc = bosh.execute("simulate", descriptor).stdout
        wInvoc = bosh.execute(
            "simulate", descriptor, "-i", bosh.example(descriptor)
        ).stdout
        self.assertEqual(noInvoc, wInvoc)

    def test_consistency_withAndWithout_invoc_withConfigFile(self):
        descriptor = self.get_file_path("test_simulate_consistency_configFile.json")
        invoc = "tmpInvoc.json"
        config = "tmpConfig.toml"
        wInvocCommand = (
            "bosh example {0}" + " > {1} " + " && bosh exec simulate {0} -i {1}"
        ).format(descriptor, invoc)
        noInvocCommand = f"bosh exec simulate {descriptor}"

        subprocess.call(wInvocCommand, shell=True)
        with open(config, "r+") as configFile:
            wInvoc = configFile.readlines()
            os.remove(config)
            os.remove(invoc)

        subprocess.call(noInvocCommand, shell=True)
        with open(config, "r+") as configFile:
            noInvoc = configFile.readlines()
            os.remove(config)

        self.assertEqual(wInvoc, noInvoc)

    def test_list_separator(self):
        ret = bosh.execute(
            "simulate",
            self.get_file_path("list_separator.json"),
            "-i",
            self.get_file_path("list_separator_inv.json"),
        )
        self.assertIn("1:2:3", ret.stdout)
        self.assertIn("foo.txt,bar.m", ret.stdout)

    def test_no_spaces(self):
        ret = bosh.execute("simulate", self.get_file_path("no_spaces.json"))
        self.assertNotIn(" ", ret.stdout)



#!/usr/bin/env python

import os

import pytest

from boutiques import bosh
from boutiques.exporter import ExportError
from boutiques.tests.BaseTest import BaseTest


class TestExport(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("export")

    def test_export(self):
        example1_desc = self.example1_descriptor
        example1_desc_doi = self.get_file_path("example1_docker_with_doi.json")
        fout = "test-example1-carmin.json"
        ref_name = "example1_docker_exported.json"
        ref_file = self.get_file_path(ref_name)
        ref_name_p2 = "example1_docker_exported_python2.json"
        ref_file_p2 = self.get_file_path(ref_name_p2)
        # Identifier is passed, descriptor has no DOI
        self.assertFalse(
            bosh(
                [
                    "export",
                    "carmin",
                    example1_desc,
                    "--identifier",
                    "123",
                    fout,
                ]
            )
        )
        result = open(fout).read().strip()
        assert (
            result == open(ref_file).read().strip()
            or result == open(ref_file_p2).read().strip()
        )
        # Identifier is not passed, descriptor has no DOI
        with self.assertRaises(ExportError) as e:
            bosh(["export", "carmin", example1_desc, fout])
        self.assertTrue(
            "Descriptor must have a DOI, or identifier "
            "must be specified" in str(e.exception)
        )
        self.assertRaises(
            ExportError,
        )
        # Identifier is not passed, descriptor has a DOI
        ref_name = "example1_docker_exported_doi.json"
        ref_file = self.get_file_path(ref_name)
        ref_name_p2 = "example1_docker_exported_doi_python2.json"
        ref_file_p2 = self.get_file_path(ref_name_p2)
        self.assertFalse(bosh(["export", "carmin", example1_desc_doi, fout]))
        result = open(fout).read().strip()
        self.assertIn(
            result,
            [open(ref_file).read().strip(), open(ref_file_p2).read().strip()],
        )
        os.remove(fout)

    def test_export_json_obj(self):
        example1_desc = open(self.example1_descriptor).read()
        example1_desc_doi = self.get_file_path("example1_docker_with_doi.json")
        fout = "test-example1-carmin.json"
        ref_name = "example1_docker_exported.json"
        ref_file = self.get_file_path(ref_name)
        ref_name_p2 = "example1_docker_exported_python2.json"
        ref_file_p2 = self.get_file_path(ref_name_p2)
        # Identifier is passed, descriptor has no DOI
        self.assertFalse(
            bosh(
                [
                    "export",
                    "carmin",
                    example1_desc,
                    "--identifier",
                    "123",
                    fout,
                ]
            )
        )
        result = open(fout).read().strip()
        self.assertIn(
            result,
            [open(ref_file).read().strip(), open(ref_file_p2).read().strip()],
        )
        # Identifier is not passed, descriptor has no DOI
        with self.assertRaises(ExportError) as e:
            bosh(["export", "carmin", example1_desc, fout])
        self.assertIn(
            "Descriptor must have a DOI, or identifier " "must be specified",
            str(e.exception),
        )
        self.assertRaises(
            ExportError,
        )
        # Identifier is not passed, descriptor has a DOI
        ref_name = "example1_docker_exported_doi.json"
        ref_file = self.get_file_path(ref_name)
        ref_name_p2 = "example1_docker_exported_doi_python2.json"
        ref_file_p2 = self.get_file_path(ref_name_p2)
        self.assertFalse(bosh(["export", "carmin", example1_desc_doi, fout]))
        result = open(fout).read().strip()
        self.assertIn(
            result,
            [open(ref_file).read().strip(), open(ref_file_p2).read().strip()],
        )
        os.remove(fout)



import json
import os
import tempfile
from unittest import mock
from urllib.request import urlopen, urlretrieve

from boutiques_mocks import MockZenodoRecord, example_boutiques_tool, mock_zenodo_search

from boutiques.bosh import bosh
from boutiques.puller import ZenodoError
from boutiques.tests.BaseTest import BaseTest


def mock_urlretrieve(*args, **kwargs):
    mock_record1 = example_boutiques_tool
    mock_record2 = MockZenodoRecord(
        2587160,
        "makeblastdb_foo",
        "",
        "https://zenodo.org/api/files/"
        "d861b2cd-ec68-4613-9847-1911904a1218/"
        "makeblastdb.json",
    )
    temp = tempfile.NamedTemporaryFile(delete=False, mode="w+t")
    if str(example_boutiques_tool.id) in args[0]:
        json.dump(mock_zenodo_search([mock_record1]).mock_json, temp)
        temp.close()
        return urlretrieve(f"file://{temp.name}", args[1])
    if "makeblastdb.json" in args[0]:
        json.dump(mock_zenodo_search([mock_record2]).mock_json, temp)
        temp.close()
        return urlretrieve(f"file://{temp.name}", args[1])
    return urlretrieve(args[0], args[1])


class TestPull(BaseTest):
    @mock.patch("boutiques.puller.urlretrieve", side_effect=mock_urlretrieve)
    def test_pull(self, mock_urlretrieve):
        bosh(["pull", "zenodo." + str(example_boutiques_tool.id)])
        cache_dir = os.path.join(
            os.path.expanduser("~"), ".cache", "boutiques", "production"
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(
                    cache_dir,
                    "zenodo-" + str(example_boutiques_tool.id) + ".json",
                )
            )
        )

    @mock.patch("boutiques.puller.urlretrieve", side_effect=mock_urlretrieve)
    def test_pull_multi(self, mock_urlretrieve):
        results = bosh(
            [
                "pull",
                "zenodo." + str(example_boutiques_tool.id),
                "zenodo.2587160",
            ]
        )
        cache_dir = os.path.join(
            os.path.expanduser("~"), ".cache", "boutiques", "production"
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(
                    cache_dir,
                    "zenodo-" + str(example_boutiques_tool.id) + ".json",
                )
            )
        )
        self.assertTrue(os.path.exists(os.path.join(cache_dir, "zenodo-2587160.json")))
        self.assertEqual(len(results), 2, results)

    @mock.patch("boutiques.puller.urlretrieve", side_effect=mock_urlretrieve)
    def test_pull_duplicate_collapses(self, mock_urlretrieve):
        results = bosh(
            [
                "pull",
                "zenodo." + str(example_boutiques_tool.id),
                "zenodo.2587160",
                "zenodo.2587160",
            ]
        )
        self.assertEqual(len(results), 2, results)

    @mock.patch("boutiques.puller.urlretrieve", side_effect=mock_urlretrieve)
    def test_pull_missing_raises_exception(self, mock_urlretrieve):
        good1 = "zenodo." + str(example_boutiques_tool.id)
        good2 = "zenodo.2587160"
        bad1 = "zenodo.9999990"
        with self.assertRaises(ZenodoError) as e:
            bosh(["pull", good1, bad1, good2])
        self.assertIn(
            f"Descriptor \"{bad1.split('.')[1]}\" not found",
            str(e.exception),
        )

    @mock.patch("boutiques.puller.urlretrieve", side_effect=mock_urlretrieve)
    def test_pull_missing_prefix(self, mock_urlretrieve):
        with self.assertRaises(ZenodoError) as e:
            bosh(["pull", str(example_boutiques_tool.id)])
        self.assertIn("Zenodo ID must be prefixed by 'zenodo'", str(e.exception))

    @mock.patch("boutiques.puller.urlretrieve", side_effect=mock_urlretrieve)
    def test_pull_not_found(self, mock_urlretrieve):
        with self.assertRaises(ZenodoError) as e:
            bosh(["pull", "zenodo.99999"])
        self.assertIn('Descriptor "99999" not found', str(e.exception))



#!/usr/bin/env python

import os
from argparse import ArgumentParser
from unittest import mock

import pytest
from boutiques_mocks import MockZenodoRecord, mock_zenodo_search

import boutiques as bosh
import boutiques.creator as bc
from boutiques import __file__ as bfile
from boutiques.tests.BaseTest import BaseTest
from boutiques.util.utils import LoadError


def mock_get(*args, **kwargs):
    query = args[0].split("=")[1]
    query = query[: query.find("&")]
    query = query.replace("*", "")

    mock_records = []
    # Return an arbitrary list of results
    for i in range(0, 10):
        mock_records.append(MockZenodoRecord(i, f"Example Tool {i}"))
    return mock_zenodo_search(mock_records)


class TestLogger(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup(
            os.path.join(os.path.dirname(bfile), "schema", "examples", "example1")
        )

    # Captures the stdout and stderr during test execution
    # and returns them as a tuple in readouterr()
    @pytest.fixture(autouse=True)
    def capture_st(self, capfd):
        self.capfd = capfd

    def test_raise_error(self):
        invocationStr = "invalid json string"
        with pytest.raises(LoadError) as e:
            bosh.execute(
                "launch",
                self.example1_descriptor,
                invocationStr,
                "--skip-data-collection",
            )
        self.assertIn("[ ERROR ]", str(e.getrepr(style="long")))

    @mock.patch("requests.get", side_effect=mock_get)
    def test_print_info(self, mock_get):
        bosh.search("-v")
        out, err = self.capfd.readouterr()
        self.assertIn("[ INFO ]", out)
        self.assertIn("[ INFO (200) ", out)

    def test_print_warning(self):
        parser = ArgumentParser(description="my tool description")
        parser.add_argument(
            "--myarg",
            "-m",
            action="store",
            help="my help",
            dest="==SUPPRESS==",
        )
        creatorObj = bc.CreateDescriptor(
            parser,
            execname="/path/to/myscript.py",
            verbose=True,
            tags={"purpose": "testing-creator", "foo": "bar"},
        )
        out, err = self.capfd.readouterr()
        self.assertIn("[ WARNING ]", out)

    def test_evaloutput(self):
        query = bosh.evaluate(
            self.example1_descriptor,
            self.get_file_path("invocation.json"),
            "invalid-query",
        )
        out, err = self.capfd.readouterr()
        self.assertIn("[ ERROR ]", out)



#!/usr/bin/env python

from boutiques.logger import print_error


def evaluateEngine(executor, query):
    try:
        # TODO: improve splitting to not fail in valid situations
        layers = query.split("/")[:-1] if "/" in query else [query]
        conditions = query.split("/")[-1] if "/" in query else []
        if len(conditions) > 0:
            if "," in conditions:
                conditions = conditions.split(",")
            else:
                conditions = [conditions]

        objs = executor.desc_dict
        for layer in layers:
            objs = objs[layer]

        query_result = {}
        for obj in objs:
            eligible = True
            for cond in conditions:
                lhs, rhs = cond.split("=")
                try:
                    rhs = float(rhs)
                except ValueError:
                    if rhs == "False":
                        rhs = False
                    elif rhs == "True":
                        rhs = True

                if obj.get(lhs) != rhs:
                    if not (obj.get(lhs) is None and rhs is False):
                        eligible = False

            if eligible:
                if "output-files" in layers:
                    query_result[obj["id"]] = executor.out_dict.get(obj["id"])
                elif "inputs" in layers:
                    # print(obj)
                    query_result[obj["id"]] = executor.in_dict.get(obj["id"])
                elif "groups" in layers:
                    query_result[obj["id"]] = {}
                    for mem in obj["members"]:
                        query_result[obj["id"]][mem] = executor.in_dict.get(mem)

        return query_result

    except Exception:
        print_error(f"Invalid query ({query}). See --help.")
        return {}



import json
import tempfile
from urllib.request import urlopen, urlretrieve

from boutiques.logger import print_info, print_warning, raise_error
from boutiques.publisher import Publisher
from boutiques.puller import Puller
from boutiques.util.utils import loadJson
from boutiques.zenodoHelper import ZenodoError, ZenodoHelper


class DeprecateError(Exception):
    pass


def deprecate(
    zenodo_id,
    by_zenodo_id=None,
    sandbox=False,
    verbose=False,
    zenodo_token=None,
    download_function=urlretrieve,
):

    # Get the descriptor and Zenodo id
    puller = Puller([zenodo_id], verbose=verbose, sandbox=sandbox)
    descriptor_fname = puller.pull()[0]
    descriptor_json = loadJson(descriptor_fname, sandbox=sandbox, verbose=verbose)

    # Return if tool is already deprecated
    deprecated = descriptor_json.get("deprecated-by-doi")
    if deprecated is not None:
        if isinstance(deprecated, str):
            print_info(f"Tool {zenodo_id} is already deprecated by {deprecated} ")
        if by_zenodo_id is not None:
            prompt = (
                "Tool {} will be deprecated by {}, "
                "this cannot be undone. Are you sure? (Y/n) "
            ).format(zenodo_id, by_zenodo_id)
            ret = input(prompt)
            if ret.upper() != "Y":
                return
        else:
            print_warning(f"Tool {zenodo_id} is already deprecated")
            return
    # Set record id and Zenodo id
    zhelper = ZenodoHelper(sandbox=sandbox, no_int=True, verbose=verbose)
    zid = zhelper.get_zid_from_filename(descriptor_fname)
    record_id = zhelper.get_record_id_from_zid(zid)

    # Return if tool has a newer version
    record = zhelper.zenodo_get_record(record_id)
    if not record["metadata"]["relations"]["version"][0]["is_last"]:
        new_version = record["metadata"]["relations"]["version"][0]["last_child"][
            "pid_value"
        ]
        raise_error(
            DeprecateError,
            "Tool {} has a newer version "
            "(zenodo.{}), it cannot be deprecated.".format(zenodo_id, new_version),
        )
        return

    # Add deprecated property
    if by_zenodo_id is None:
        descriptor_json["deprecated-by-doi"] = True
    else:
        # Check that by_zenodo_id exists
        by_record_id = zhelper.get_record_id_from_zid(by_zenodo_id)
        if zhelper.record_exists(by_record_id) is False:
            raise_error(DeprecateError, f"Tool does not exist: {by_zenodo_id}")
        # Assign deprecated-by property
        by_doi_id = zhelper.get_doi_from_zid(by_zenodo_id)
        descriptor_json["deprecated-by-doi"] = by_doi_id

    # Add doi to descriptor (mandatory for update)
    if descriptor_json.get("doi") is None:
        descriptor_json["doi"] = zhelper.get_doi_from_zid(zid)

    # Save descriptor in temp file
    tmp = tempfile.NamedTemporaryFile(delete=False, mode="w", suffix=".json")
    content = json.dumps(descriptor_json, indent=4, sort_keys=True)
    tmp.write(content)
    tmp.close()

    # Publish updated descriptor
    publisher = Publisher(
        tmp.name,
        zenodo_token,
        replace=True,
        sandbox=sandbox,
        no_int=True,
        id="zenodo." + zid,
        verbose=verbose,
    )
    return publisher.publish()



#!/usr/bin/env python

import boutiques
from boutiques.tests.BaseTest import BaseTest


class TestBoutiquesTools(BaseTest):

    def test_python_interface_validate(self):
        self.setup("invocation")
        fil = self.get_file_path("good.json")
        self.assertIsNone(boutiques.validate(fil))



import os
from collections import OrderedDict

import simplejson as json

from boutiques import __file__ as bfile
from boutiques.logger import print_warning, raise_error


# Utility function to wrap modules that use non-essential libs
def importCatcher():
    def decorate(f):
        def applicator(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except ImportError as e:
                print(e)
                raise ImportError(
                    "Try installing the full version of Boutiques "
                    "with: \n\t pip install -e boutiques[all]"
                )

        return applicator

    return decorate


# Parses absolute path into filename
def extractFileName(path):
    # Helps OS path handle case where "/" is at the end of path
    if path is None:
        return None
    elif path[:-1] == "/":
        return os.path.basename(path[:-1]) + "/"
    else:
        return os.path.basename(path)


class LoadError(Exception):
    pass


# Helper function that loads the JSON object coming from either a string,
# a local file or a file pulled from Zenodo
def loadJson(userInput, verbose=False, sandbox=False):
    # Check for JSON file (local or from Zenodo)
    json_file = None
    if os.path.isfile(userInput):
        json_file = userInput
    elif "zenodo" in ".".join(userInput.split(".")[:-1]).lower():
        from boutiques.puller import Puller

        puller = Puller([userInput], verbose, sandbox)
        json_file = puller.pull()[0]
    if json_file is not None:
        with open(json_file) as f:
            return OrderedDict(json.loads(f.read(), object_pairs_hook=OrderedDict))
    # JSON file not found, so try to parse JSON object
    e = (
        "Cannot parse input {}: file not found, "
        "invalid Zenodo ID, or invalid JSON object"
    ).format(userInput)
    if userInput.isdigit():
        raise_error(LoadError, e)
    try:
        return OrderedDict(json.loads(userInput, object_pairs_hook=OrderedDict))
    except ValueError:
        raise_error(LoadError, e)


# Helper function that takes a conditional path template key as input,
# and outputs a formatted string that isolates variables/values from
# operators, parentheses, and python keywords with a space.
# ex: "(opt1>2)" becomes " ( opt1 > 2 ) "
# "(opt1<=10.1)" becomes " ( opt1 <= 10.1 ) "
def conditionalExpFormat(s):
    cleanedExpression = ""
    idx = 0
    while idx < len(s):
        c = s[idx]
        if c in ["=", "!", "<", ">"]:
            cleanedExpression += f" {c}{'=' if s[idx + 1] == '=' else ' '}"
            idx += 1
        elif c in ["(", ")"]:
            cleanedExpression += f" {c} "
        else:
            cleanedExpression += c
        idx += 1
    return cleanedExpression


# Sorts and returns a descriptor dictionary according to
# the keys' order in a template descriptor
def customSortDescriptorByKey(
    descriptor,
    template=os.path.join(
        os.path.dirname(bfile), "templates", "ordered_keys_desc.json"
    ),
):

    def sortListedObjects(objList, template):
        sortedObjList = []
        for obj in objList:
            sortedObj = OrderedDict()
            for key in template:
                if key in obj:
                    sortedObj[key] = obj[key]
            for key in obj:
                if key not in template:
                    sortedObj[key] = obj[key]
            sortedObjList.append(sortedObj)

        if len(objList) != len(sortedObjList):
            print_warning("Sorted list does not represent" " original list.")
            return objList
        for obj, sobj in zip(objList, sortedObjList):
            if obj != dict(sobj):
                print_warning("Sorted list does not represent" " original list.")
                return objList
        return sortedObjList

    template = loadJson(template)
    sortedDesc = OrderedDict()

    # Add k:v to sortedDesc according to their order in template
    for key in template:
        if key in descriptor:
            if type(descriptor[key]) is list:
                sortedDesc[key] = sortListedObjects(descriptor[key], template[key][0])
            elif type(descriptor[key]) is dict:
                sortedDesc[key] = customSortDescriptorByKey(
                    descriptor[key], template=json.dumps(template[key])
                )
            else:
                sortedDesc[key] = descriptor[key]

    # Add remaining k:v that are missing from template
    for key in descriptor:
        if key not in sortedDesc:
            sortedDesc[key] = descriptor[key]
    if sortedDesc != descriptor:
        print_warning("Sorted descriptor does not represent" " original descriptor.")
        return descriptor
    return sortedDesc


# Sorts tool invocations according to descriptor's inputs'
def customSortInvocationByInput(invocation, descriptor):
    descriptor = loadJson(descriptor)
    # sort invoc according to input's order in descriptor
    sortedInvoc = OrderedDict()
    sortedInvoc.update(
        {
            key: invocation[key]
            for key in [
                inp["id"]
                for inp in descriptor["inputs"]
                if descriptor["inputs"] is not None
            ]
            if key in invocation
        }
    )
    if sortedInvoc != invocation:
        print_warning("Sorted invocation does not represent" " original invocation.")
        return invocation
    return sortedInvoc


def snakeCaseToCamelCase(id):
    words = id.split("_")
    for idx, word in enumerate(words[1:]):
        if word[0].islower():
            words[idx + 1] = word[0].upper() + word[1:]
    return "".join(words)


def camelCaseInputIds(descriptor):
    conversion_dict = {}
    if "inputs" in descriptor:
        for inp in descriptor["inputs"]:
            camelCaseId = snakeCaseToCamelCase(inp["id"])
            conversion_dict[inp["id"]] = camelCaseId

    # Find all instances of old input ids
    # and replace them with camelCase ids
    plainTextDesc = json.dumps(descriptor, indent=2)
    for k, v in conversion_dict.items():
        # Only replace ids surrounded by single/double quotes,
        # in case the the old input ids are used in other strings
        plainTextDesc = plainTextDesc.replace(f'"{k}"', f'"{v}"')
        plainTextDesc = plainTextDesc.replace(f"'{k}'", f"'{v}'")
    descriptor = json.loads(plainTextDesc)
    return descriptor


def formatSphinxUsage(func, usage_str):
    args = (
        usage_str.replace("[", " ").replace("]", " ").replace("\n", "").split(func)[1:]
    )
    args = "".join(args)
    args = args.split("  ")[0:]
    args = list(filter(lambda x: x != "", args))
    args = [f'"{arg.strip()}"' for arg in args]
    args = ", ".join(args)
    return f"[{args}]"



import json

from boutiques import execute, pprint, validate
from boutiques.util.utils import loadJson


def function(descriptor):
    """
    Returns a function to invoke bosh.execute on a descriptor.
    args:
        descriptor: Zenodo id, file name, or JSON string representing a
                    descriptor.
        name: name of the function to create. Defaults to the tool name in the
              descriptor.
    """

    validate(descriptor)
    descriptor_json = loadJson(descriptor)

    def f(*args, **kwargs):

        # Set default mode to 'launch'
        if len(args) > 0:
            mode = args[0]
        else:
            mode = "launch"
        if mode not in ["launch", "simulate"]:
            mode = "launch"
        else:
            args = args[1:]

        # Call bosh execute
        if mode == "launch":
            return execute(mode, descriptor, json.dumps(kwargs), *args)
        if len(kwargs) > 0:
            return execute(mode, descriptor, "-i", json.dumps(kwargs), *args)
        return execute(mode, descriptor, *args)

    f.__name__ = str(descriptor_json["name"])

    # Documentation
    doc = []
    doc.append(
        r"""Runs {} through its Boutiques interface.
    *args:
        - mode: 'launch' or 'simulate'. Defaults to 'launch'.
        - other arguments: will be passed to bosh execute. Examples: '-s',
          '-x'. See help(bosh.execute) for a complete list.
    *kwargs:
        {} arguments as defined in the Boutiques descriptor, referenced
        from input ids. Example: {}='some_value'. See complete
        list in descriptor help below.

""".format(
            f.__name__, f.__name__, descriptor_json["inputs"][0]["id"]
        )
    )
    doc.append(pprint(descriptor))
    f.__doc__ = "".join(doc)

    return f



#!/usr/bin/env python

import os.path as op
import re

import simplejson

from boutiques.logger import print_info, raise_error
from boutiques.validator import DescriptorValidationError


# BIDS validation module
def validate_bids(descriptor, valid=False):

    if not valid:
        msg = "Please provide a Boutiques descriptor that has been validated."
        raise_error(DescriptorValidationError, msg)

    errors = []

    # TODO: verify not only that all fields/keys exist, their properties, too

    # Ensure the command-line conforms to the BIDS app spec
    msg_template = "   CLIError: command-line doesn't match template: {}"
    cltemp = (
        r"mkdir -p \[OUTPUT_DIR\]; (.*) \[BIDS_DIR\] \[OUTPUT_DIR\]"
        r" \[ANALYSIS_LEVEL\] \[PARTICIPANT_LABEL\] \[SESSION_LABEL\]"
        r"[\\s]*(.*)"
    )
    cmdline = descriptor["command-line"]
    if len(re.findall(cltemp, cmdline)) < 1:
        errors += [msg_template.format(cltemp)]

    # Verify IDs are present which link to the OUTPUT_DIR
    # key bot as File and String
    ftypes = {"File", "String"}
    msg_template = '   OutError: "{}" types for outdir do not match "{}"'
    outtypes = {
        inp["type"]
        for inp in descriptor["inputs"]
        if inp["value-key"] == "[OUTPUT_DIR]"
    }
    if outtypes != ftypes:
        errors += [msg_template.format(", ".join(outtypes), ", ".join(ftypes))]

    # Verify that analysis levels is an enumerable with some
    # subset of "participant", "session", and "group"
    choices = ["session", "participant", "group"]
    msg_template = ' LevelError: "{}" is not a valid analysis level'
    alevels = [
        inp["value-choices"]
        for inp in descriptor["inputs"]
        if inp["value-key"] == "[ANALYSIS_LEVEL]"
    ][0]
    errors += [msg_template.format(lv) for lv in alevels if lv not in choices]

    # Verify there is only a single output defined (the directory)
    msg_template = "OutputError: 0 or multiple outputs defined"
    if len(descriptor["output-files"]) != 1:
        errors += [msg_template]
    else:
        # Verify that the output shows up as an output
        msg_template = "OutputError: OUTPUT_DIR is not represented as an output"
        if descriptor["output-files"][0]["path-template"] != "[OUTPUT_DIR]":
            errors += [msg_template]

    errors = None if errors == [] else errors
    if errors is None:
        print_info("BIDS validation OK")
    else:
        raise_error(
            DescriptorValidationError,
            "Invalid BIDS app descriptor:" "\n" + "\n".join(errors),
        )



import subprocess

import simplejson as json

import boutiques


def CreateBoshDocs():
    # Get file as plain text
    def readFile(path):
        with open(path) as file:
            return file.read()

    # create an RST index page
    def createDocsPage(path, docString):
        with open(path, "w+") as page:
            page.write(docString)

    # Start CLI docs .rst files creation
    indexTemplate = readFile("./_templates/index.rst")
    cliTemplate = readFile("./_templates/cli.rst")
    pythonAPITemplate = readFile("./_templates/pythonApi.rst")
    # Split readme into two, slicing at command line api section
    readmeText = readFile("../README.rst")

    indexDocString = readmeText + "\n" + indexTemplate + "\n"

    # Create a RST page per api
    createDocsPage("cli.rst", cliTemplate)
    createDocsPage("pythonApi.rst", pythonAPITemplate)
    createDocsPage("index.rst", indexDocString)


CreateBoshDocs()
subprocess.Popen(["make", "clean"], cwd="./").wait()
make = subprocess.Popen(["make", "html"], cwd="./").wait()



#!/usr/bin/env python

# A script for handling invocation schema generation and
# validation for Boutiques application descriptors
# Requires jsonschema 2.5

import argparse
import os
import sys
from functools import reduce

import jsonschema
from jsonschema import ValidationError

from boutiques.logger import print_info, raise_error


# An exception class specific to invocations
class InvocationValidationError(ValidationError):
    pass


# Generate an invocation schema from a Boutiques application descriptor
def generateInvocationSchema(toolDesc, oname=None, validateWrtMetaSchema=True):
    # Subclass dictionary to act more like a ruby hash map
    class RMap(dict):
        def __getitem__(self, key):
            return dict.get(self, key)

    s = RMap(toolDesc)
    descName, inputs, groups = s["name"], s["inputs"], s["groups"] or []
    oname = descName + ".invocationSchema" if oname is None else oname
    schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "title": str(oname),
        "description": "Invocation schema for " + str(descName.lower()) + ".",
        "type": "object",
        "additionalProperties": False,
    }
    if inputs is None:
        return schema

    # Handle type constraints for single inputs
    def addTypeConstraints(h, ind):
        i = RMap(ind)
        id, type, target = i["id"], i["type"], None

        def undertype():
            if i["integer"]:
                return {"type": "integer"}
            elif i["value-choices"]:
                return {"enum": i["value-choices"]}
            elif type == "File":
                return {"type": "string"}
            elif type == "Flag":
                return {"type": "boolean"}
            else:
                return {"type": type.lower()}

        if i["list"]:
            h[id] = {"type": "array", "items": undertype()}
            if i["min-list-entries"]:
                h[id]["minItems"] = i["min-list-entries"]
            if i["max-list-entries"]:
                h[id]["maxItems"] = i["max-list-entries"]
            target = h[id]["items"]
        else:
            h[id] = undertype()
            target = h[id]
        if not i["minimum"] is None:
            target["minimum"] = i["minimum"]  # Python treats 0 as false
        if not i["maximum"] is None:
            target["maximum"] = i["maximum"]
        if i["exclusive-minimum"]:
            target["exclusiveMinimum"] = i["exclusive-minimum"]
        if i["exclusive-maximum"]:
            target["exclusiveMaximum"] = i["exclusive-maximum"]
        return h

    schema["properties"] = reduce(addTypeConstraints, inputs, {})
    # Required inputs
    reqInputs = [x["id"] for x in [x for x in inputs if not x.get("optional")]]
    if len(list(reqInputs)) > 0:
        schema["required"] = reqInputs

    # Helper functions (assumes properly formed descriptor)
    def byInd(id):
        if not [i for i in inputs if id == i["id"]]:
            return [i for i in groups if id == i["id"]][0]
        return [i for i in inputs if id == i["id"]][0]

    def isFlag(id):
        return "type" in byInd(id) and byInd(id)["type"] == "Flag"

    # Handle one-is-required groups
    g1r = [g for g in groups if g.get("one-is-required")]

    def reqMember(m):  # Generate code for a one-is-required member
        if isFlag(m):
            return {"properties": {m: {"enum": [True]}}, "required": [m]}
        else:
            return {"required": [m]}

    if len(g1r) > 0:
        schema["allOf"] = [{"anyOf": list(map(reqMember, g["members"]))} for g in g1r]

    # Handle requires and disables-inputs/mutex constraints
    def handleDisablesRequires(h, inval):
        i, h = RMap(inval), RMap(h)
        id, grps = i["id"], {g["id"]: g for g in groups}
        reqs = [req for req in i["requires-inputs"]] if "requires-inputs" in i else []
        disbs = i["disables-inputs"] or []
        # Mutex group members added to disablees list
        mutex_group_members = [
            g for g in groups if "mutually-exclusive" in g and (id in g["members"])
        ]
        for g in mutex_group_members:
            for m in [m for m in g["members"] if not m == id]:
                disbs.append(m)
        # Handle requires-inputs (flags have to be true to satisfy it)
        if reqs:
            reqMap = {}
            for r in reqs:
                if isFlag(r):
                    reqMap[r] = {"enum": [True]}
            if isFlag(id):
                h[id] = {}
                h[id]["anyOf"] = [
                    {"properties": reqMap, "required": reqs},
                    {"properties": {id: {"enum": [False]}}},
                ]
            else:
                h[id] = {}
                # Handle requiring mutex group
                for r in [r for r in reqs if r in grps]:
                    # Group properties will be tested in validator
                    reqs.remove(r)
                    h[id]["anyOf"] = []
                    for m in grps[r]["members"]:
                        reqMap[m] = schema["properties"][m]
                        h[id]["anyOf"].append(
                            {"properties": reqMap, "required": reqs + [m]}
                        )
                h[id]["properties"] = reqMap
                if reqs:
                    h[id]["required"] = reqs
        # Handle disables-inputs (false flags do not violate
        # the disabled criteria)
        if len(disbs) > 0:
            if h[id] is None:
                h[id] = RMap({})

            def makeSingleDisablesMap(dh, d):  # Flags can be absent or false
                if isFlag(d):
                    dh[d] = {"oneOf": [{"not": {}}, {"enum": [False]}]}
                else:
                    dh[d] = {"not": {}}
                return dh

            disbMap = reduce(makeSingleDisablesMap, disbs, {})
            if isFlag(id):  # Flags do not disable their targets when false
                if h[id].get("anyOf") is None:
                    h[id]["anyOf"] = [
                        {"properties": disbMap},
                        {"properties": {id: {"enum": [False]}}},
                    ]
                # Need to add disables properties to requires ones,
                # if the latter exist
                else:
                    h[id]["anyOf"][0]["properties"].update(disbMap)
            else:
                if h[id].get("properties") is None:
                    h[id]["properties"] = RMap({})
                h[id]["properties"].update(disbMap)
        return h

    schema["dependencies"] = reduce(handleDisablesRequires, inputs, {})
    # Validate wrt meta-schema if desired
    if validateWrtMetaSchema:
        validateSchema(schema)
    # Return final schema
    return schema


# Validate data with respect to the invocation schema
def validateSchema(s, d=None, **kwargs):
    # Check schema wrt meta-schema
    try:
        jsonschema.Draft4Validator.check_schema(s)
    except jsonschema.SchemaError as se:
        errExit("Invocation schema is invalid.\n" + str(se.message), False)
    # Check data instance against schema
    if d:
        try:
            jsonschema.validate(d, s)
        except ValidationError as e:
            raise_error(InvocationValidationError, e)
        if kwargs.get("verbose"):
            print_info("Invocation Schema validation OK")


# Script exit helper
def errExit(msg, parser, err_code=1):
    sys.stderr.write("Error: " + msg + "\n")
    sys.exit(err_code)
